"""Tests Grid classes."""

import numpy as np
import pytest
from scipy import special

from desc.equilibrium import Equilibrium
from desc.examples import get
from desc.grid import (
    ConcentricGrid,
    Grid,
    LinearGrid,
    QuadratureGrid,
    dec_to_cf,
    find_least_rational_surfaces,
    find_most_rational_surfaces,
    most_rational,
)
from desc.profiles import PowerSeriesProfile


class TestGrid:
    """Test for Grid classes."""

    @pytest.mark.unit
    def test_linear_grid(self):
        """Test node placement in a LinearGrid."""
        L, M, N, NFP, axis, endpoint = 8, 5, 3, 2, True, False
        g = LinearGrid(L, M, N, NFP, sym=False, axis=axis, endpoint=endpoint)

        np.testing.assert_equal(g.num_rho, L + 1)
        np.testing.assert_equal(g.num_theta, 2 * M + 1)
        np.testing.assert_equal(g.num_zeta, 2 * N + 1)
        assert g.endpoint == endpoint

        nodes = np.stack(
            [
                np.tile(
                    np.repeat(np.linspace(1, 0, g.num_rho, axis)[::-1], g.num_theta),
                    g.num_zeta,
                ),
                np.tile(
                    np.linspace(0, 2 * np.pi, g.num_theta, endpoint),
                    g.num_rho * g.num_zeta,
                ),
                np.repeat(
                    np.linspace(0, 2 * np.pi / NFP, g.num_zeta, endpoint),
                    g.num_rho * g.num_theta,
                ),
            ]
        ).T

        np.testing.assert_allclose(g.nodes, nodes)
        np.testing.assert_allclose(g.weights.sum(), (2 * np.pi) ** 2)
        # spacing.prod != weights for grid with duplicates
        if not endpoint:
            np.testing.assert_allclose(g.spacing.prod(axis=1), g.weights)

    @pytest.mark.unit
    def test_grid_LMN(self):
        """Make sure grid attributes LMN are set correctly."""
        theta = 5
        g1 = LinearGrid(theta=theta, zeta=3, rho=2, sym=True)
        assert g1.num_theta == (theta + 1) // 2
        g2 = LinearGrid(M=g1.M, N=g1.N, L=g1.L, sym=True)
        assert (g1.L, g1.M, g1.N) == (g2.L, g2.M, g2.N)
        g1 = LinearGrid(theta=theta, zeta=3, rho=2, sym=False)
        assert g1.num_theta == theta
        g2 = LinearGrid(M=g1.M, N=g1.N, L=g1.L, sym=False)
        assert (g1.L, g1.M, g1.N) == (g2.L, g2.M, g2.N)

    @pytest.mark.unit
    def test_linear_grid_spacing_consistency(self):
        """Test consistency between alternate construction methods."""

        def test(sym, endpoint, axis, ntheta_is_odd):
            nrho = 1
            ntheta = 6 + ntheta_is_odd
            nzeta = 7
            NFP = 3
            # We need to construct the theta coordinates to match
            # whatever is done in grid.py for construction via theta=integer input.
            # First round up ntheta + endpoint to next even integer.
            theta = np.linspace(
                0,
                2 * np.pi,
                (
                    (ntheta + endpoint + 1) // 2 * 2 - endpoint
                    if (sym and ntheta > 1)
                    else ntheta
                ),
                endpoint=endpoint,
            )
            if sym and ntheta > 1:
                # Second, change the array to match the effect done in grid.py,
                # which sets dtheta to a constant. This requires shifting nodes
                # off the axis.
                theta += theta[1] / 2
                # Third, manually delete last node because the addition
                # above pushed it to the (2pi, pi) range which = (0, pi)
                # range, meaning it wouldn't have been deleted by
                # enforce_symmetry which only removes nodes from (pi, 2pi).
                theta = theta[: theta.size - endpoint]
            lg_1 = LinearGrid(
                rho=nrho,
                theta=ntheta,
                zeta=nzeta,
                NFP=NFP,
                sym=sym,
                axis=axis,
                endpoint=endpoint,
            )
            lg_2 = LinearGrid(
                rho=np.linspace(1, 0, nrho, endpoint=axis)[::-1],
                theta=theta,
                zeta=np.linspace(0, 2 * np.pi / NFP, nzeta, endpoint=endpoint),
                NFP=NFP,
                sym=sym,
                axis=axis,
                endpoint=endpoint,
            )
            np.testing.assert_allclose(lg_1.nodes, lg_2.nodes)
            np.testing.assert_allclose(lg_1.spacing, lg_2.spacing)
            np.testing.assert_allclose(lg_1.weights, lg_2.weights)

        # ntheta is odd
        test(sym=False, endpoint=False, axis=False, ntheta_is_odd=True)
        test(sym=False, endpoint=False, axis=True, ntheta_is_odd=True)
        test(sym=False, endpoint=True, axis=False, ntheta_is_odd=True)
        test(sym=False, endpoint=True, axis=True, ntheta_is_odd=True)
        test(sym=True, endpoint=False, axis=False, ntheta_is_odd=True)
        test(sym=True, endpoint=False, axis=True, ntheta_is_odd=True)
        test(sym=True, endpoint=True, axis=False, ntheta_is_odd=True)
        test(sym=True, endpoint=True, axis=True, ntheta_is_odd=True)
        # ntheta is even
        test(sym=False, endpoint=False, axis=False, ntheta_is_odd=False)
        test(sym=False, endpoint=False, axis=True, ntheta_is_odd=False)
        test(sym=False, endpoint=True, axis=False, ntheta_is_odd=False)
        test(sym=False, endpoint=True, axis=True, ntheta_is_odd=False)
        test(sym=True, endpoint=False, axis=False, ntheta_is_odd=False)
        test(sym=True, endpoint=False, axis=True, ntheta_is_odd=False)
        test(sym=True, endpoint=True, axis=False, ntheta_is_odd=False)
        test(sym=True, endpoint=True, axis=True, ntheta_is_odd=False)

    @pytest.mark.unit
    def test_linear_grid_symmetric_nodes_consistency(self):
        """Test that specifying theta nodes from [0, pi] is sufficient."""
        # uniform spacing
        theta = np.linspace(0, 2 * np.pi, 10)
        lg_1 = LinearGrid(L=5, theta=theta, N=5, sym=True)
        lg_2 = LinearGrid(L=5, theta=theta[theta <= np.pi], N=5, sym=True)
        np.testing.assert_allclose(lg_1.nodes, lg_2.nodes)
        np.testing.assert_allclose(lg_1.spacing, lg_2.spacing)
        np.testing.assert_allclose(lg_1.weights, lg_2.weights)

        # non-uniform spacing
        pts = np.linspace(0, 1, 10)
        rho = np.array([r**3 for r in pts])
        theta = 2 * np.pi * np.array([t**1.85 for t in pts])
        zeta = 2 * np.pi * np.array([z**4 for z in pts])
        lg_1 = LinearGrid(rho=rho, theta=theta, zeta=zeta, sym=True)
        lg_2 = LinearGrid(rho=rho, theta=theta[theta <= np.pi], zeta=zeta, sym=True)
        np.testing.assert_allclose(lg_1.nodes, lg_2.nodes)
        np.testing.assert_allclose(lg_1.spacing, lg_2.spacing)
        np.testing.assert_allclose(lg_1.weights, lg_2.weights)

    @pytest.mark.unit
    def test_linear_grid_spacing_two_nodes(self):
        """Test that 2 node grids assign equal spacing to nodes."""
        node_count = 2
        NFP = 7  # any integer > 1 is good candidate for test

        def test(endpoint):
            lg = LinearGrid(
                theta=np.linspace(0, 2 * np.pi, node_count, endpoint=endpoint),
                zeta=np.linspace(0, 2 * np.pi / NFP, node_count, endpoint=endpoint),
                NFP=NFP,
                sym=False,
                endpoint=endpoint,
            )
            # When endpoint is true the rho nodes spacing should be scaled down
            # so that theta and zeta surface integrals weigh the duplicate
            # nodes less.
            spacing = np.tile(
                [1 / 2 if endpoint else 1, np.pi, np.pi], (node_count**2, 1)
            )
            np.testing.assert_allclose(lg.spacing, spacing)

        test(endpoint=False)
        test(endpoint=True)

    @pytest.mark.unit
    def test_spacing_when_duplicate_node_is_removed(self):
        """Test grid spacing when the duplicate node is removed due to symmetry."""
        sym = True
        endpoint = True
        spacing = np.tile([1, np.pi, 2 * np.pi], (2, 1))

        lg = LinearGrid(L=0, M=1, N=0, sym=sym, endpoint=endpoint)
        np.testing.assert_allclose(lg.spacing, spacing)

        lg_2 = LinearGrid(
            rho=np.linspace(1, 0, num=1)[::-1],
            theta=np.linspace(0, 2 * np.pi, num=3, endpoint=endpoint),
            zeta=np.linspace(0, 2 * np.pi, num=1, endpoint=endpoint),
            sym=sym,
            endpoint=endpoint,
        )
        np.testing.assert_allclose(lg_2.spacing, spacing)
        np.testing.assert_allclose(lg_2.weights, spacing.prod(axis=1))

    @pytest.mark.unit
    def test_node_spacing_non_sym(self):
        """Test surface spacing on grids with sym=False."""
        self._test_node_spacing_non_sym(False, 8, 13, 3)
        self._test_node_spacing_non_sym(True, 8, 13, 3)
        self._test_node_spacing_non_sym(False, 1, 1, 3)

    @staticmethod
    def _test_node_spacing_non_sym(
        endpoint, unique_theta_count, unique_zeta_count, NFP
    ):
        """Test surface spacing on grids with sym=False."""
        nrho = 1
        sym = False

        def test(grid):
            if not endpoint:
                # if no duplicates
                np.testing.assert_allclose(grid.weights, grid.spacing.prod(axis=1))
            is_theta_dupe = (endpoint and unique_theta_count > 0 and not sym) & (
                grid.nodes[:, 1] % (2 * np.pi) == 0
            )
            is_zeta_dupe = (endpoint and unique_zeta_count > 0) & (
                grid.nodes[:, 2] % (2 * np.pi / NFP) == 0
            )

            def test_surface(surface_label, actual_ds, desired_ds):
                for index, ds in enumerate(actual_ds):
                    if is_theta_dupe[index] and is_zeta_dupe[index]:
                        # the grid has 4 of these nodes
                        np.testing.assert_allclose(
                            ds, desired_ds / 4, err_msg=surface_label
                        )
                    elif is_theta_dupe[index] or is_zeta_dupe[index]:
                        # the grid has 2 of these nodes
                        np.testing.assert_allclose(
                            ds, desired_ds / 2, err_msg=surface_label
                        )
                    else:
                        # unique node
                        np.testing.assert_allclose(
                            ds, desired_ds, err_msg=surface_label
                        )

            test_surface(
                "rho",
                grid.spacing[:, 1:].prod(axis=1),
                (2 * np.pi / max(unique_theta_count, 1))
                * (2 * np.pi / max(unique_zeta_count, 1)),
            )
            test_surface(
                "theta",
                grid.spacing[:, [0, 2]].prod(axis=1),
                (1 / nrho) * (2 * np.pi / max(unique_zeta_count, 1)),
            )
            test_surface(
                "zeta",
                grid.spacing[:, :2].prod(axis=1),
                (1 / nrho) * (2 * np.pi / max(unique_theta_count, 1)),
            )

        lg_1 = LinearGrid(
            rho=nrho,
            theta=unique_theta_count + endpoint,
            zeta=unique_zeta_count + endpoint,
            NFP=NFP,
            sym=sym,
            endpoint=endpoint,
        )
        rho = np.linspace(1, 0, nrho)[::-1]
        theta = np.linspace(
            0, 2 * np.pi, unique_theta_count + endpoint, endpoint=endpoint
        )
        zeta = np.linspace(
            0, 2 * np.pi / NFP, unique_zeta_count + endpoint, endpoint=endpoint
        )
        lg_2 = LinearGrid(
            rho=rho, theta=theta, zeta=zeta, NFP=NFP, sym=sym, endpoint=endpoint
        )
        lg_3 = LinearGrid(
            rho=rho,
            theta=theta,
            zeta=zeta,
            NFP=NFP,
            sym=sym,
            endpoint=not endpoint,  # incorrect marker should have no effect
        )
        assert lg_3.endpoint == endpoint
        test(lg_1)
        # The test below might fail for theta and zeta surfaces only if nrho > 1.
        # This is unrelated to how duplicate node spacing is handled.
        # The cause is because grid construction does not always
        # compute drho as constant (even when the rho nodes are linearly spaced),
        # and this test assumes drho to be a constant for grids without duplicates.
        test(lg_2)
        test(lg_3)

    @pytest.mark.unit
    def test_symmetry_spacing_basic(self):
        """Test symmetry effect on spacing in a basic case."""
        nrho = 2
        ntheta = 3
        nzeta = 2
        lg = LinearGrid(rho=nrho, theta=ntheta, zeta=nzeta, sym=False)
        lg_sym = LinearGrid(rho=nrho, theta=ntheta, zeta=nzeta, sym=True)
        np.testing.assert_allclose(
            lg.spacing,
            np.tile(
                [1 / nrho, 2 * np.pi / ntheta, 2 * np.pi / nzeta],
                (nrho * ntheta * nzeta, 1),
            ),
        )
        np.testing.assert_allclose(lg.weights, lg.spacing.prod(axis=1))
        np.testing.assert_allclose(
            lg_sym.spacing,
            np.tile(
                [1 / nrho, 2 * np.pi / (ntheta - 1), 2 * np.pi / nzeta],
                (nrho * (ntheta - 1) * nzeta, 1),
            ),
        )
        np.testing.assert_allclose(lg_sym.weights, lg_sym.spacing.prod(axis=1))

    @pytest.mark.unit
    def test_node_spacing_sym(self):
        """Test surface spacing on grids with sym=True."""
        # It's useful to test with ntheta even and odd.
        # These tests should pass for any combination of parameters.
        self._test_node_spacing_sym(False, 8, 13, 3)
        self._test_node_spacing_sym(False, 7, 13, 3)
        self._test_node_spacing_sym(True, 8, 13, 3)
        self._test_node_spacing_sym(True, 7, 13, 3)

    @staticmethod
    def _test_node_spacing_sym(endpoint, ntheta, unique_zeta_count, NFP):
        """Test surface spacing on grids with sym=True."""
        nrho = 1
        sym = True

        # unique_theta_and_reflection count is twice
        # the number of nodes at unique theta coordinates (0, pi)
        # on a given theta curve.
        def test(grid, unique_theta_and_reflection_count):
            is_zeta_dupe = (endpoint and unique_zeta_count > 0) & (
                grid.nodes[:, 2] % (2 * np.pi / NFP) == 0
            )

            def test_surface(label, actual_ds, desired_ds):
                for index, ds in enumerate(actual_ds):
                    if label != "theta" and sym and grid.nodes[index, 1] % np.pi != 0:
                        # these nodes should have double weight to account for
                        # reflection across symmetry line
                        if is_zeta_dupe[index]:
                            # the grid has 2 of these nodes,
                            # so each should have half weight
                            np.testing.assert_allclose(
                                ds, (desired_ds / 2) * 2, err_msg=label
                            )
                        else:
                            np.testing.assert_allclose(
                                ds, desired_ds * 2, err_msg=label
                            )
                    elif is_zeta_dupe[index]:
                        # the grid has 2 of these nodes,
                        # so each should have half weight
                        np.testing.assert_allclose(ds, desired_ds / 2, err_msg=label)
                    else:
                        # unique node
                        np.testing.assert_allclose(ds, desired_ds, err_msg=label)

            test_surface(
                "rho",
                grid.spacing[:, 1:].prod(axis=1),
                (2 * np.pi / max(unique_theta_and_reflection_count, 1))
                * (2 * np.pi / max(unique_zeta_count, 1)),
            )
            test_surface(
                "theta",
                grid.spacing[:, [0, 2]].prod(axis=1),
                (1 / nrho) * (2 * np.pi / max(unique_zeta_count, 1)),
            )
            test_surface(
                "zeta",
                grid.spacing[:, :2].prod(axis=1),
                (1 / nrho) * (2 * np.pi / max(unique_theta_and_reflection_count, 1)),
            )

        lg_1_sym = LinearGrid(
            rho=nrho,
            theta=ntheta,
            zeta=unique_zeta_count + endpoint,
            NFP=NFP,
            sym=sym,
            endpoint=endpoint,
        )
        # Recall that LinearGrid created with integers forces ntheta + endpoint
        # to be even and shifts nodes counterclockwise. The result is that
        # unique_theta_and_reflection_count is ntheta rounded up (down) to the
        # closest even integer when endpoint is false (true).
        test(
            lg_1_sym, unique_theta_and_reflection_count=(ntheta - endpoint + 1) // 2 * 2
        )

        rho = np.linspace(1, 0, nrho)[::-1]
        theta = np.linspace(0, 2 * np.pi, ntheta, endpoint=endpoint)
        zeta = np.linspace(
            0, 2 * np.pi / NFP, unique_zeta_count + endpoint, endpoint=endpoint
        )
        lg_2_sym = LinearGrid(
            rho=rho, theta=theta, zeta=zeta, NFP=NFP, sym=sym, endpoint=endpoint
        )
        lg_3_sym = LinearGrid(
            rho=rho,
            theta=theta,
            zeta=zeta,
            NFP=NFP,
            sym=sym,
            endpoint=not endpoint,  # incorrect marker should have no effect
        )
        # endpoints were deleted
        assert not lg_1_sym.endpoint
        assert not lg_2_sym.endpoint
        assert not lg_3_sym.endpoint
        # The test below might fail for theta and zeta surfaces only if nrho > 1.
        # This is unrelated to how duplicate node spacing is handled.
        # The cause is because grid construction does not always
        # compute drho as constant (even when the rho nodes are linearly spaced),
        # and this test assumes drho to be a constant for grids without duplicates.
        test(lg_2_sym, unique_theta_and_reflection_count=ntheta - endpoint)
        test(lg_3_sym, unique_theta_and_reflection_count=ntheta - endpoint)

    @pytest.mark.unit
    def test_concentric_grid(self):
        """Test node placement in ConcentricGrid."""
        M = 2
        N = 0
        NFP = 1
        grid_ansi = ConcentricGrid(
            M, M, N, NFP, sym=False, axis=True, node_pattern="linear"
        )
        grid_fringe = ConcentricGrid(
            2 * M, M, N, NFP, sym=True, axis=True, node_pattern="linear"
        )
        ansi_nodes = np.stack(
            [
                np.array([0, 1, 1, 1, 1, 1]),
                np.array(
                    [0, 0, 2 * np.pi / 5, 4 * np.pi / 5, 6 * np.pi / 5, 8 * np.pi / 5]
                ),
                np.zeros((int((M + 1) * (M + 2) / 2),)),
            ]
        ).T
        fringe_nodes = np.stack(
            [
                np.array([0, 0.5, 0.5, 1, 1, 1]),
                np.array(
                    [
                        np.pi / 2,
                        np.pi / 4,
                        3 * np.pi / 4,
                        np.pi / 6,
                        np.pi / 2,
                        5 * np.pi / 6,
                    ]
                ),
                np.zeros((6,)),
            ]
        ).T
        np.testing.assert_allclose(grid_ansi.nodes, ansi_nodes, err_msg="ansi")
        np.testing.assert_allclose(grid_fringe.nodes, fringe_nodes, err_msg="fringe")
        np.testing.assert_allclose(grid_ansi.weights.sum(), (2 * np.pi) ** 2)

    @pytest.mark.unit
    def test_concentric_grid_high_res(self):
        """Test that we can create high resolution grids without crashing.

        Test for GH issue #207.
        """
        _ = ConcentricGrid(L=32, M=28, N=30)

    @pytest.mark.unit
    def test_quadrature_grid(self):
        """Test node placement in QuadratureGrid."""
        L = 2
        M = 2
        N = 0
        NFP = 1
        grid_quad = QuadratureGrid(L, M, N, NFP)
        roots, weights = special.js_roots(2, 2, 2)
        quadrature_nodes = np.stack(
            [
                np.array(
                    [roots[0]] * grid_quad.num_theta + [roots[1]] * grid_quad.num_theta
                ),
                np.array(
                    [0, 2 * np.pi / 5, 4 * np.pi / 5, 6 * np.pi / 5, 8 * np.pi / 5] * 2
                ),
                np.zeros(10),
            ]
        ).T
        np.testing.assert_allclose(grid_quad.spacing.prod(axis=1), grid_quad.weights)
        np.testing.assert_allclose(grid_quad.nodes, quadrature_nodes)

    @pytest.mark.unit
    def test_quad_grid_volume_integration(self):
        """Test that quadrature grid gives correct volume integrals."""
        r = 1
        R = 10
        vol = 2 * (np.pi**2) * (r**2) * R

        inputs = {
            "L": 1,
            "M": 1,
            "N": 0,
            "NFP": 1,
            "Psi": 1.0,
            "pressure": np.array([[0, 0]]),
            "iota": np.array([[0, 0]]),
            "surface": np.array([[0, 0, 0, R, 0], [0, 1, 0, r, 0], [0, -1, 0, 0, -r]]),
            "spectral_indexing": "ansi",
        }

        eq = Equilibrium(**inputs)
        grid = QuadratureGrid(L=eq.L, M=eq.M, N=eq.N, NFP=eq.NFP)
        g = eq.compute("sqrt(g)", grid=grid)
        vol_quad = np.sum(np.abs(g["sqrt(g)"]) * grid.weights)

        np.testing.assert_allclose(vol, vol_quad)

    @pytest.mark.unit
    def test_repr(self):
        """Test string representations of grid objects."""
        qg = ConcentricGrid(2, 3, 4)
        s = str(qg)
        assert "ConcentricGrid" in s
        assert "jacobi" in s
        assert "L=2" in s
        assert "M=3" in s
        assert "N=4" in s

    @pytest.mark.unit
    def test_change_resolution(self):
        """Test changing grid resolution."""

        def test(grid, *desired_resolution):
            assert (grid.L, grid.M, grid.N, grid.NFP) == desired_resolution
            assert grid.num_rho == grid.unique_rho_idx.size
            assert grid.num_theta == grid.unique_theta_idx.size
            assert grid.num_zeta == grid.unique_zeta_idx.size
            np.testing.assert_equal(
                (grid.unique_rho_idx, grid.inverse_rho_idx),
                np.unique(grid.nodes[:, 0], return_index=True, return_inverse=True)[1:],
            )
            np.testing.assert_equal(
                (grid.unique_theta_idx, grid.inverse_theta_idx),
                np.unique(grid.nodes[:, 1], return_index=True, return_inverse=True)[1:],
            )
            np.testing.assert_equal(
                (grid.unique_zeta_idx, grid.inverse_zeta_idx),
                np.unique(grid.nodes[:, 2], return_index=True, return_inverse=True)[1:],
            )
            np.testing.assert_array_equal(
                grid.axis, np.nonzero(grid.nodes[:, 0] == 0)[0]
            )
            # test that changing NFP updated the nodes
            assert np.isclose(
                grid.nodes[grid.unique_zeta_idx[-1], 2],
                (grid.num_zeta - 1) / grid.num_zeta * 2 * np.pi / grid.NFP,
            )

        lg = LinearGrid(1, 2, 3)
        lg.change_resolution(2, 3, 4, 5)
        test(lg, 2, 3, 4, 5)
        qg = QuadratureGrid(1, 2, 3)
        qg.change_resolution(2, 3, 4, 5)
        test(qg, 2, 3, 4, 5)
        cg = ConcentricGrid(2, 3, 4)
        cg.change_resolution(3, 4, 5, 2)
        test(cg, 3, 4, 5, 2)
        cg = ConcentricGrid(2, 3, 4)
        cg.change_resolution(cg.L, cg.M, cg.N, NFP=5)
        test(cg, cg.L, cg.M, cg.N, 5)

    @pytest.mark.unit
    def test_enforce_symmetry(self):
        """Test correctness of enforce_symmetry() for uniformly spaced nodes.

        Unlike enforce_symmetry(), the algorithm used in LinearGrid for
        symmetry also works if the nodes are not uniformly spaced. This test
        compares the two methods when the grid is uniformly spaced in theta,
        as a means to ensure enforce_symmetry() is correct.
        """
        ntheta = 6
        theta = np.linspace(0, 2 * np.pi, ntheta, endpoint=False)
        lg_1 = LinearGrid(L=5, theta=theta, N=4, NFP=4, sym=True)
        lg_2 = LinearGrid(L=5, theta=theta, N=4, NFP=4, sym=False)
        # precondition for the following tests to work
        np.testing.assert_allclose(lg_2.spacing[:, 1], 2 * np.pi / ntheta)

        lg_2._sym = True
        lg_2._enforce_symmetry()
        np.testing.assert_allclose(lg_1.nodes, lg_2.nodes)
        np.testing.assert_allclose(lg_1.spacing, lg_2.spacing)
        lg_2._weights = lg_2._scale_weights()
        np.testing.assert_allclose(lg_1.spacing, lg_2.spacing)
        np.testing.assert_allclose(lg_1.weights, lg_2.weights)

    @pytest.mark.unit
    def test_symmetry_volume_integral(self):
        """Test volume integral of a symmetric function."""
        L = [3, 3, 5, 3]
        M = [3, 6, 5, 7]
        N = [2, 2, 2, 2]
        NFP = [5, 3, 5, 3]
        sym = np.array([True, True, False, False])
        # to test code not tested on grids made with M=.
        even_number = 4
        n_theta = even_number - sym

        # Currently, midpoint rule is false for LinearGrid made with L=number.
        def test(grid, midpoint_rule=False):
            r = grid.nodes[:, 0]
            t = grid.nodes[:, 1]
            z = grid.nodes[:, 2] * grid.NFP
            true_surface_avg = 5
            function_of_rho = 1 / (r + 0.35)
            f = (
                true_surface_avg
                + np.cos(t)
                - 0.5 * np.cos(z)
                + 3 * np.cos(t) * np.cos(z) ** 2
                - 2 * np.sin(z) * np.sin(t)
            ) * function_of_rho

            if midpoint_rule:
                r_unique = r[grid.unique_rho_idx]
                dr = np.empty_like(r_unique)
                dr[0] = (r_unique[0] + r_unique[1]) / 2
                dr[1:-1] = (r_unique[2:] - r_unique[:-2]) / 2
                dr[-1] = 1 - (r_unique[-2] + r_unique[-1]) / 2
            else:
                dr = 1 / grid.num_rho
            expected_integral = np.sum(dr * function_of_rho[grid.unique_rho_idx])
            true_integral = np.log(1.35 / 0.35)
            midpoint_rule_error_bound = np.max(dr) ** 2 / 24 * (2 / 0.35**3)
            right_riemann_error_bound = dr * (1 / 0.35 - 1 / 1.35)
            np.testing.assert_allclose(
                expected_integral,
                true_integral,
                rtol=0,
                atol=(
                    midpoint_rule_error_bound / 4
                    if midpoint_rule
                    else right_riemann_error_bound / 3
                ),
                err_msg=type(grid),
            )
            np.testing.assert_allclose(
                np.sum(grid.weights * f) / (4 * np.pi**2 * true_surface_avg),
                expected_integral,
                rtol=1e-15,
                err_msg=type(grid),
            )

        for i in range(len(L)):
            test(LinearGrid(L=L[i], M=M[i], N=N[i], NFP=NFP[i], sym=sym[i]))
            test(LinearGrid(L=L[i], theta=n_theta[i], N=N[i], NFP=NFP[i], sym=sym[i]))
            test(
                LinearGrid(
                    L=L[i],
                    theta=np.linspace(0, 2 * np.pi, n_theta[i], endpoint=False),
                    N=N[i],
                    NFP=NFP[i],
                    sym=sym[i],
                )
            )
            test(
                LinearGrid(
                    L=L[i],
                    theta=np.linspace(0, 2 * np.pi, n_theta[i] + 1, endpoint=False),
                    N=N[i],
                    NFP=NFP[i],
                    sym=sym[i],
                )
            )
            test(
                ConcentricGrid(L=L[i], M=M[i], N=N[i], NFP=NFP[i], sym=sym[i]),
                midpoint_rule=True,
            )
            # nonuniform spacing when sym is False, but spacing is still symmetric
            test(
                LinearGrid(
                    L=L[i],
                    theta=np.linspace(0, np.pi, n_theta[i]),
                    N=N[i],
                    NFP=NFP[i],
                    sym=sym[i],
                )
            )
            test(
                LinearGrid(
                    L=L[i],
                    theta=np.linspace(0, np.pi, n_theta[i] + 1),
                    N=N[i],
                    NFP=NFP[i],
                    sym=sym[i],
                )
            )

    @pytest.mark.unit
    def test_compress_expand_inverse_op(self):
        """Test that compress & expand are inverse operations for surface functions.

        Each test should be done on different types of grids
        (e.g. LinearGrid, ConcentricGrid) and grids with duplicate nodes
        (e.g. endpoint=True).
        """

        def test(surface_label, grid):
            r = np.random.random_sample(
                size={
                    "rho": grid.num_rho,
                    "theta": grid.num_theta,
                    "zeta": grid.num_zeta,
                }[surface_label]
            )
            expanded = grid.expand(r, surface_label)
            assert expanded.size == grid.num_nodes
            s = grid.compress(expanded, surface_label)
            np.testing.assert_allclose(r, s, err_msg=surface_label)

        L, M, N, NFP = 6, 6, 3, 5
        lg_endpoint = LinearGrid(L=L, M=M, N=N, NFP=NFP, sym=True, endpoint=True)
        cg_sym = ConcentricGrid(L=L, M=M, N=N, NFP=NFP, sym=True)
        test("rho", lg_endpoint)
        test("theta", lg_endpoint)
        test("zeta", lg_endpoint)
        test("rho", cg_sym)
        test("theta", cg_sym)
        test("zeta", cg_sym)

    @pytest.mark.unit
    def test_meshgrid(self):
        """Test meshgrid constructor."""
        R = np.linspace(0, 1, 4)
        A = np.linspace(0, 2 * np.pi, 2)
        Z = np.linspace(0, 2 * np.pi, 3)
        grid = Grid.create_meshgrid(
            [R, A, Z], coordinates="raz", period=(np.inf, 2 * np.pi, 2 * np.pi)
        )
        # treating theta == alpha just for grid construction
        grid1 = LinearGrid(rho=R, theta=A, zeta=Z)
        # atol=1e-12 bc Grid by default shifts points away from the axis a tiny bit
        np.testing.assert_allclose(grid1.nodes, grid.nodes, atol=1e-12)
        # want radial/poloidal/toroidal nodes sorted in the same order for both
        np.testing.assert_allclose(grid1.unique_rho_idx, grid.unique_rho_idx)
        np.testing.assert_allclose(grid1.unique_theta_idx, grid.unique_alpha_idx)
        np.testing.assert_allclose(grid1.unique_zeta_idx, grid.unique_zeta_idx)
        np.testing.assert_allclose(grid1.inverse_rho_idx, grid.inverse_rho_idx)
        np.testing.assert_allclose(grid1.inverse_theta_idx, grid.inverse_alpha_idx)
        np.testing.assert_allclose(grid1.inverse_zeta_idx, grid.inverse_zeta_idx)

    @pytest.mark.unit
    def test_meshgrid_reshape(self):
        """Test that reshaping meshgrids works correctly."""
        grid = LinearGrid(2, 3, 4)

        r = grid.nodes[grid.unique_rho_idx, 0]
        t = grid.nodes[grid.unique_theta_idx, 1]
        z = grid.nodes[grid.unique_zeta_idx, 2]

        # user regular allclose for broadcasting to work correctly
        # reshaping rtz should have rho along first axis
        assert np.allclose(
            grid.meshgrid_reshape(grid.nodes[:, 0], "rtz"), r[:, None, None]
        )
        # reshaping rzt should have theta along last axis
        assert np.allclose(
            grid.meshgrid_reshape(grid.nodes[:, 1], "rzt"), t[None, None, :]
        )
        # reshaping tzr should have zeta along 2nd axis
        assert np.allclose(
            grid.meshgrid_reshape(grid.nodes, "tzr")[:, :, :, 2], z[None, :, None]
        )

        # coordinates are rtz, not raz
        with pytest.raises(ValueError):
            grid.meshgrid_reshape(grid.nodes[:, 0], "raz")

        # not a meshgrid
        grid = ConcentricGrid(2, 3, 4)
        with pytest.raises(ValueError):
            grid.meshgrid_reshape(grid.nodes[:, 0], "rtz")

        rho = np.linspace(0, 1, 3)
        alpha = np.linspace(0, 2 * np.pi, 4)
        zeta = np.linspace(0, 6 * np.pi, 5)
        grid = Grid.create_meshgrid([rho, alpha, zeta], coordinates="raz")
        r, a, z = grid.nodes.T
        # functions of zeta should separate along first two axes
        # since those are contiguous, this should work
        f = grid.meshgrid_reshape(z, "raz").reshape(-1, zeta.size)
        for i in range(1, f.shape[0]):
            np.testing.assert_allclose(f[i - 1], f[i])
        # likewise for rho
        f = grid.meshgrid_reshape(r, "raz").reshape(rho.size, -1)
        for i in range(1, f.shape[-1]):
            np.testing.assert_allclose(f[:, i - 1], f[:, i])
        # test reshaping result won't mix data
        f = grid.meshgrid_reshape(a**2 + z, "raz")
        for i in range(1, f.shape[0]):
            np.testing.assert_allclose(f[i - 1], f[i])
        f = grid.meshgrid_reshape(r**2 + z, "raz")
        for i in range(1, f.shape[1]):
            np.testing.assert_allclose(f[:, i - 1], f[:, i])
        f = grid.meshgrid_reshape(r**2 + a, "raz")
        for i in range(1, f.shape[-1]):
            np.testing.assert_allclose(f[..., i - 1], f[..., i])


@pytest.mark.unit
def test_find_most_rational_surfaces():
    """Test finding the most rational surfaces and their locations."""
    # simple test, linear iota going from 1 to 3
    iota = PowerSeriesProfile([1, 2])
    rho, io = find_most_rational_surfaces(iota, 5)
    np.testing.assert_allclose(rho, np.linspace(0, 1, 5), atol=1e-14, rtol=0)
    np.testing.assert_allclose(io, np.linspace(1, 3, 5), atol=1e-14, rtol=0)

    # simple test, linear iota going from -1 to -3
    iota = PowerSeriesProfile([-1, -2])
    rho, io = find_most_rational_surfaces(iota, 5)
    np.testing.assert_allclose(rho, np.linspace(0, 1, 5), atol=1e-14, rtol=0)
    np.testing.assert_allclose(io, np.linspace(-1, -3, 5), atol=1e-14, rtol=0)

    # invalid (a > b) ranges and negative ranges are swapped and made positive
    all_same = [
        most_rational(1, 2),
        most_rational(2, 1),
        most_rational(-1, -2),
        most_rational(-2, -1),
    ]

    assert len(set(map(abs, all_same))) == 1

    # if 0 in range, return 0
    has_zero = [
        most_rational(0, 1),
        most_rational(0, -1),
        most_rational(0, 0),
        most_rational(-1, 0),
        most_rational(1, 0),
    ]

    assert all(result == 0 for result in has_zero)


@pytest.mark.unit
def test_find_least_rational_surfaces():
    """Test finding the least rational surfaces and their locations."""
    # simple test, linear iota going from 1 to 3
    iota = PowerSeriesProfile([1, 2])
    rhor, ior = find_most_rational_surfaces(iota, 10)
    rho, io = find_least_rational_surfaces(iota, 10)
    # to compare how rational/irrational things are, we use the length of the
    # continued fraction. Not a great test due to rounding errors, but seems to work
    lio = [len(dec_to_cf(i)) for i in io]
    lior = [len(dec_to_cf(i)) for i in ior]
    max_rational = max(lior)

    assert np.all(np.array(lio) > max_rational)


@pytest.mark.unit
def test_custom_jitable_grid_indexing():
    """Test that unique/inverse indices are set correctly when jitable=True."""
    eq = get("NCSX")
    with pytest.warns(UserWarning, match="Reducing radial"):
        eq.change_resolution(3, 3, 3, 6, 6, 6)
    # field lines on two surfaces
    rho = np.concatenate([0.5 * np.ones(10), 0.7 * np.ones(10)])
    theta = np.concatenate([np.linspace(0, 1, 10), np.linspace(0, 1, 10)]) * 2 * np.pi
    zeta = np.concatenate([np.linspace(0, 1, 10), np.linspace(0, 1, 10)]) * 2 * np.pi
    grid1 = Grid(np.array([rho, theta, zeta]).T, jitable=False)
    grid2 = Grid(np.array([rho, theta, zeta]).T, jitable=True)
    grid3 = Grid(np.array([rho, theta, zeta]).T, jitable=True, _unique_rho_idx=[0, 10])
    np.testing.assert_allclose(grid1.nodes, grid2.nodes)

    x = np.random.random(grid1.num_nodes)

    # these shouldn't error
    _ = grid1.unique_rho_idx
    _ = grid1.unique_theta_idx
    _ = grid1.unique_zeta_idx
    _ = grid1.inverse_rho_idx
    _ = grid1.inverse_theta_idx
    _ = grid1.inverse_zeta_idx
    _ = grid3.unique_rho_idx

    with pytest.raises(AttributeError):
        _ = grid2.unique_rho_idx
    with pytest.raises(AttributeError):
        _ = grid2.unique_theta_idx
    with pytest.raises(AttributeError):
        _ = grid2.unique_zeta_idx
    with pytest.raises(AttributeError):
        _ = grid2.inverse_rho_idx
    with pytest.raises(AttributeError):
        _ = grid2.inverse_theta_idx
    with pytest.raises(AttributeError):
        _ = grid2.inverse_zeta_idx

    assert not hasattr(grid1, "weights")
    assert not hasattr(grid1, "spacing")
    assert not hasattr(grid1, "source_grid")
    assert not hasattr(grid2, "num_rho")
    assert not hasattr(grid2, "num_theta")
    assert not hasattr(grid2, "num_zeta")

    y1 = grid1.copy_data_from_other(x, grid2, "rho")
    y2 = grid2.copy_data_from_other(x, grid1, "rho")
    y3 = grid3.copy_data_from_other(x, grid1, "rho")

    np.testing.assert_allclose(y1, y2)
    np.testing.assert_allclose(y1, y3)

    # make sure compress/expand done when override_grid=True works as expected
    b1 = eq.compute(["|B|"], grid=grid1, override_grid=True)["|B|"]
    with pytest.raises(AttributeError):
        _ = eq.compute(["|B|"], grid=grid2, override_grid=True)["|B|"]
    b3 = eq.compute(["|B|"], grid=grid3, override_grid=True)["|B|"]
    np.testing.assert_allclose(b1, b3)



"""High order method for singular surface integrals, from Malhotra 2019."""

from abc import ABC, abstractmethod

import numpy as np
import scipy
from interpax import fft_interp2d

from desc.backend import fori_loop, jnp, put, vmap
from desc.basis import DoubleFourierSeries
from desc.compute.geom_utils import rpz2xyz, rpz2xyz_vec, xyz2rpz_vec
from desc.grid import LinearGrid
from desc.io import IOAble
from desc.utils import isalmostequal, islinspaced, safediv, safenorm


def _get_quadrature_nodes(q):
    """Polar nodes for quadrature around singular point.

    Parameters
    ----------
    q : int
        Order of quadrature in radial and azimuthal directions.

    Returns
    -------
    r, w : ndarray
        Radial and azimuthal coordinates
    dr, dw : ndarray
        Radial and azimuthal spacing/quadrature weights

    """
    Nr = Nw = q
    r, dr = scipy.special.roots_legendre(Nr)
    # integrate separately over [-1,0] and [0,1]
    r1 = 1 / 2 * r - 1 / 2
    r2 = 1 / 2 * r + 1 / 2
    r = jnp.concatenate([r1, r2])
    dr = jnp.concatenate([dr, dr]) / 2
    w = np.linspace(0, np.pi, Nw, endpoint=False)
    dw = jnp.ones_like(w) * np.pi / Nw
    r, w = jnp.meshgrid(r, w)
    r = r.flatten()
    w = w.flatten()
    dr, dw = jnp.meshgrid(dr, dw)
    dr = dr.flatten()
    dw = dw.flatten()
    return r, w, dr, dw


class _BIESTInterpolator(IOAble, ABC):
    """Base class for interpolators from cartesian to polar domain.

    Used for singular integral calculations.

    Parameters
    ----------
    eval_grid, source_grid : Grid
        Evaluation and source points for the integral transform.
        source_grid should be a LinearGrid
    s : int
        Extent of polar grid in number of source grid points. Same as "M" in the
        original Malhotra papers.
    q : int
        Order of quadrature in polar domain

    """

    _io_attrs_ = ["_eval_grid", "_source_grid", "_q", "_s"]

    @abstractmethod
    def __init__(self, eval_grid, source_grid, s, q):
        pass

    @property
    def s(self):
        """int: Extent of polar grid in number of source grid points."""
        return self._s

    @property
    def q(self):
        """int: Order of quadrature in polar domain."""
        return self._q

    @abstractmethod
    def __call__(self, f, i):
        """Interpolate data to polar grid points.

        Parameters
        ----------
        f : ndarray
            Data at source grid points to interpolate.
        i : int
            Index of polar node.

        Returns
        -------
        fi : ndarray
            Source data interpolated to ith polar node.
        """


class FFTInterpolator(_BIESTInterpolator):
    """FFT interpolation operator required for high order singular integration.

    Parameters
    ----------
    eval_grid, source_grid : Grid
        Evaluation and source points for the integral transform.
        Both should be LinearGrid without stellarator symmetry.
    s : int
        Extent of polar grid in number of source grid points. Same as "M" in the
        original Malhotra papers.
    q : int
        Order of quadrature in polar domain

    """

    _io_attrs_ = _BIESTInterpolator._io_attrs_ + ["_h_t", "_h_z", "_st", "_sz"]

    def __init__(self, eval_grid, source_grid, s, q):
        # current fft interpolating can't handle symmetric grids correctly
        assert not source_grid.sym, "fft interpolator requires non-symmetric grid"
        assert not eval_grid.sym, "fft interpolator requires non-symmetric grid"

        # need source_grid to be linearly spaced in theta, zeta,
        # and contain only 1 rho value
        assert isalmostequal(
            source_grid.nodes[:, 0]
        ), "singular integration requires source grid on a single surface"
        assert source_grid.num_nodes == (
            source_grid.num_theta * source_grid.num_zeta
        ), "singular integration requires a tensor product grid in theta and zeta"
        source_theta = source_grid.nodes[:, 1].reshape(
            (source_grid.num_zeta, source_grid.num_theta)
        )
        source_zeta = source_grid.nodes[:, 2].reshape(
            (source_grid.num_zeta, source_grid.num_theta)
        )
        assert isalmostequal(
            source_theta, axis=0
        ), "singular integration requires rectangular source grid in theta and zeta"
        assert isalmostequal(
            source_zeta, axis=1
        ), "singular integration requires rectangular source grid in theta and zeta"
        assert islinspaced(
            source_theta, axis=1
        ), "singular integration requires source nodes be equally spaced in theta"
        assert islinspaced(
            source_zeta, axis=0
        ), "singular integration requires source nodes be equally spaced in zeta"

        # need eval_grid to be linearly spaced in theta, zeta,
        # and contain only 1 rho value
        assert isalmostequal(
            eval_grid.nodes[:, 0]
        ), "singular integration requires eval grid on a single surface"
        assert np.all(source_grid.nodes[:, 0] == eval_grid.nodes[0, 0]) and np.all(
            eval_grid.nodes[:, 0] == source_grid.nodes[0, 0]
        ), "singular integration requires source and eval grids on the same surface."
        assert eval_grid.num_nodes == (
            eval_grid.num_theta * eval_grid.num_zeta
        ), "singular integration requires a tensor product grid in theta and zeta"
        eval_theta = eval_grid.nodes[:, 1].reshape(
            (eval_grid.num_zeta, eval_grid.num_theta)
        )
        eval_zeta = eval_grid.nodes[:, 2].reshape(
            (eval_grid.num_zeta, eval_grid.num_theta)
        )
        assert isalmostequal(
            eval_theta, axis=0
        ), "singular integration requires rectangular eval grid in theta and zeta"
        assert isalmostequal(
            eval_zeta, axis=1
        ), "singular integration requires rectangular eval grid in theta and zeta"
        assert islinspaced(
            eval_theta, axis=1
        ), "singular integration requires eval nodes be equally spaced in theta"
        assert islinspaced(
            eval_zeta, axis=0
        ), "singular integration requires eval nodes be equally spaced in zeta"

        self._eval_grid = eval_grid
        self._source_grid = source_grid
        self._s = s
        self._q = q

        r, w, dr, dw = _get_quadrature_nodes(q)

        source_dtheta = source_grid.spacing[:, 1]
        source_dzeta = source_grid.spacing[:, 2] / source_grid.NFP

        self._h_t = jnp.mean(source_dtheta)
        self._h_z = jnp.mean(source_dzeta)

        self._st = s / 2 * self._h_t * r * jnp.sin(w)
        self._sz = s / 2 * self._h_z * r * jnp.cos(w)

    def __call__(self, f, i):
        """Interpolate data to polar grid points.

        Parameters
        ----------
        f : ndarray
            Data at source grid points to interpolate.
        i : int
            Index of polar node.

        Returns
        -------
        fi : ndarray
            Source data interpolated to ith polar node.
        """
        shp = f.shape[1:]
        f = f.reshape(
            (self._source_grid.num_theta, self._source_grid.num_zeta, -1), order="F"
        )
        g = fft_interp2d(
            f,
            self._eval_grid.num_theta,
            self._eval_grid.num_zeta,
            sx=self._st[i],
            sy=self._sz[i],
            dx=self._h_t,
            dy=self._h_z,
        )
        return g.reshape((self._eval_grid.num_nodes, *shp), order="F")


class DFTInterpolator(_BIESTInterpolator):
    """Fourier interpolation matrix required for high order singular integration.

    Parameters
    ----------
    eval_grid, source_grid : Grid
        Evaluation and source points for the integral transform.
        source_grid should be a LinearGrid without stellarator symmetry.
    s : int
        Extent of polar grid in number of source grid points. Same as "M" in the
        original Malhotra papers.
    q : int
        Order of quadrature in polar domain

    """

    _io_attrs_ = _BIESTInterpolator._io_attrs_ + ["_mat"]

    def __init__(self, eval_grid, source_grid, s, q):
        # need source_grid to be linearly spaced in theta, zeta,
        # and contain only 1 rho value
        assert isalmostequal(
            source_grid.nodes[:, 0]
        ), "singular integration requires source grid on a single surface"
        assert source_grid.num_nodes == (
            source_grid.num_theta * source_grid.num_zeta
        ), "singular integration requires a tensor product grid in theta and zeta"
        source_theta = source_grid.nodes[:, 1].reshape(
            (source_grid.num_zeta, source_grid.num_theta)
        )
        source_zeta = source_grid.nodes[:, 2].reshape(
            (source_grid.num_zeta, source_grid.num_theta)
        )
        assert isalmostequal(
            source_theta, axis=0
        ), "singular integration requires rectangular source grid in theta and zeta"
        assert isalmostequal(
            source_zeta, axis=1
        ), "singular integration requires rectangular source grid in theta and zeta"
        assert islinspaced(
            source_theta, axis=1
        ), "singular integration requires source nodes be equally spaced in theta"
        assert islinspaced(
            source_zeta, axis=0
        ), "singular integration requires source nodes be equally spaced in zeta"
        assert np.all(source_grid.nodes[:, 0] == eval_grid.nodes[0, 0]) and np.all(
            eval_grid.nodes[:, 0] == source_grid.nodes[0, 0]
        ), "singular integration requires source and eval grids on the same surface."

        self._eval_grid = eval_grid
        self._source_grid = source_grid
        self._s = s
        self._q = q

        r, w, dr, dw = _get_quadrature_nodes(q)

        source_dtheta = source_grid.spacing[:, 1]
        source_dzeta = source_grid.spacing[:, 2] / source_grid.NFP
        eval_theta = eval_grid.nodes[:, 1]
        eval_zeta = eval_grid.nodes[:, 2]

        h_t = jnp.mean(source_dtheta)
        h_z = jnp.mean(source_dzeta)

        theta_q = eval_theta[:, None] + s / 2 * h_t * r * jnp.sin(w)
        zeta_q = eval_zeta[:, None] + s / 2 * h_z * r * jnp.cos(w)

        basis = DoubleFourierSeries(
            M=source_grid.M, N=source_grid.N, NFP=source_grid.NFP
        )
        A = basis.evaluate(source_grid.nodes)
        Ainv = jnp.linalg.pinv(A)

        B = jnp.zeros((*theta_q.shape, basis.num_modes))

        def body(i, B):
            x = jnp.array(
                [
                    jnp.zeros_like(theta_q[i]),
                    theta_q[i],
                    zeta_q[i],
                ]
            ).T
            Bi = basis.evaluate(jnp.atleast_2d(x))
            B = put(B, i, Bi)
            return B

        B = fori_loop(0, B.shape[0], body, B)
        self._mat = B @ Ainv

    def __call__(self, f, i):
        """Interpolate data to polar grid points.

        Parameters
        ----------
        f : ndarray
            Data at source grid points to interpolate.
        i : int
            Index of polar node.

        Returns
        -------
        fi : ndarray
            Source data interpolated to ith polar node.
        """
        return self._mat[:, i] @ f


def _chi(rho):
    """Partition of unity function. Eq 39 in [2]."""
    return jnp.exp(-36 * jnp.abs(rho) ** 8)


def _rho(theta, zeta, theta0, zeta0, dtheta, dzeta, s):
    """Polar grid radial coordinate. Argument of Chi in eq. 36 in [2]."""
    dt = abs(theta - theta0)
    dz = abs(zeta - zeta0)
    dt = jnp.minimum(dt, 2 * np.pi - dt)
    dz = jnp.minimum(dz, 2 * np.pi - dz)
    return 2 / s * jnp.sqrt((dt / dtheta) ** 2 + (dz / dzeta) ** 2)


def _nonsingular_part(
    eval_data, eval_grid, source_data, source_grid, s, kernel, loop=False
):
    """Integrate kernel over non-singular points.

    Generally follows sec 3.2.1 of [2].
    """
    assert source_grid.NFP == int(source_grid.NFP)
    source_theta = source_grid.nodes[:, 1]
    source_zeta = source_grid.nodes[:, 2]
    source_dtheta = source_grid.spacing[:, 1]
    source_dzeta = source_grid.spacing[:, 2] / source_grid.NFP
    eval_theta = jnp.asarray(eval_grid.nodes[:, 1])
    eval_zeta = jnp.asarray(eval_grid.nodes[:, 2])
    w = source_grid.weights * source_data["|e_theta x e_zeta|"] / source_grid.NFP
    h_t = jnp.mean(source_dtheta)
    h_z = jnp.mean(source_dzeta)

    source_phi = source_data["phi"]
    keys = kernel.keys

    def nfp_loop(j, f_data):
        # calculate effects at all eval pts from all source pts on a single field
        # period, summing over field periods
        f, source_data = f_data
        source_data["zeta"] = (source_zeta + j * 2 * np.pi / source_grid.NFP) % (
            2 * np.pi
        )
        source_data["phi"] = (source_phi + j * 2 * np.pi / source_grid.NFP) % (
            2 * np.pi
        )

        # nest this def to avoid having to pass the modified source_data around the loop
        # easier to just close over it and let JAX figure it out
        def eval_pt_vmap(i):
            # this calculates the effect at a single evaluation point, from all others
            # in a single field period. vmap this to get all pts
            k = kernel(
                {key: val[i] for key, val in eval_data.items() if key in keys},
                source_data,
            ).reshape((1, source_grid.num_nodes, kernel.ndim))

            rho = _rho(
                source_theta,
                source_data["zeta"],  # to account for different field periods
                eval_theta[i],
                eval_zeta[i],
                h_t,
                h_z,
                s,
            )

            eta = _chi(rho)  # from eq 36 of [2]
            k = (1 - eta)[None, :, None] * k
            f_temp = jnp.sum(k * w[None, :, None], axis=1)
            return f_temp

        def eval_pt_loop(i, fj):
            # this calculates the effect at a single evaluation point, from all others
            # in a single field period. loop this to get all pts
            f_temp = eval_pt_vmap(i)
            return put(fj, i, f_temp.reshape(fj[i].shape))

        # vmap for inner part found more efficient than fori_loop, especially on gpu,
        # but for jacobian looped seems to be better and less memory
        if loop:
            fj = fori_loop(0, eval_grid.num_nodes, eval_pt_loop, jnp.zeros_like(f))
        else:
            fj = vmap(eval_pt_vmap)(jnp.arange(eval_grid.num_nodes))

        f += fj.reshape((eval_grid.num_nodes, kernel.ndim))
        return f, source_data

    f = jnp.zeros((eval_grid.num_nodes, kernel.ndim))
    f, _ = fori_loop(0, int(source_grid.NFP), nfp_loop, (f, source_data))

    # undo rotation of source_zeta
    source_data["zeta"] = source_zeta
    source_data["phi"] = source_phi
    # we sum distance vectors, so they need to be in xyz for that to work
    # but then need to convert vectors back to rpz
    if kernel.ndim == 3:
        f = xyz2rpz_vec(f, phi=eval_data["phi"])
    return f


def _singular_part(
    eval_data,
    eval_grid,
    source_data,
    source_grid,
    s,
    q,
    kernel,
    interpolator,
    loop=False,
):
    """Integrate singular point by interpolating to polar grid.

    Generally follows sec 3.2.2 of [2], with the following differences:

    - hyperparameter M replaced by s
    - density sigma / function f is absorbed into kernel.
    """
    source_dtheta = source_grid.spacing[:, 1]
    source_dzeta = source_grid.spacing[:, 2] / source_grid.NFP
    eval_theta = jnp.asarray(eval_grid.nodes[:, 1])
    eval_zeta = jnp.asarray(eval_grid.nodes[:, 2])

    r, w, dr, dw = _get_quadrature_nodes(q)

    h_t = jnp.mean(source_dtheta)
    h_z = jnp.mean(source_dzeta)

    eta = _chi(r)
    # integrand of eq 38 in [2] except stuff that needs to be interpolated
    v = eta * s**2 * h_t * h_z / 4 * abs(r) * dr * dw
    keys = list(set(["|e_theta x e_zeta|"] + kernel.keys))
    if "phi" in keys:
        keys += ["omega"]
    fsource = [source_data[key] for key in keys]

    def polar_pt_vmap(i):
        # evaluate the effect from a single polar node around each eval point
        # on that eval point. Polar grids from other singularities have no effect
        # See sec 3.2.2 of [2]
        dt = s / 2 * h_t * r[i] * jnp.sin(w[i])
        dz = s / 2 * h_z * r[i] * jnp.cos(w[i])
        theta_i = eval_theta + dt
        zeta_i = eval_zeta + dz

        # data interpolated to each eval pt offset by dt,dz
        source_data_polar = {
            key: interpolator(val, i) for key, val in zip(keys, fsource)
        }

        # can't interpolate phi directly since its not periodic, so we interpolate
        # omega and add it back in
        source_data_polar["zeta"] = zeta_i
        source_data_polar["theta"] = theta_i
        if "phi" in keys:
            source_data_polar["phi"] = zeta_i + source_data_polar["omega"]

        # eval pts x source pts for 1 polar grid offset
        # only need diagonal term because polar grid points
        # don't contribute to other eval pts
        k = kernel(
            {key: val for key, val in eval_data.items() if key in keys},
            source_data_polar,
            diag=True,
        ).reshape((eval_grid.num_nodes, kernel.ndim))

        dS = (v[i] * source_data_polar["|e_theta x e_zeta|"])[:, None]
        fi = k * dS
        return fi

    def polar_pt_loop(i, f):
        # this calculates the effect at a single evaluation point, from all others
        # in a single field period. loop this to get all pts
        f_temp = polar_pt_vmap(i)
        return f + f_temp.reshape((eval_grid.num_nodes, kernel.ndim))

    f = jnp.zeros((eval_grid.num_nodes, kernel.ndim))
    # vmap found more efficient than fori_loop, esp on gpu, but uses more memory
    if loop:
        f = fori_loop(0, v.size, polar_pt_loop, f)
    else:
        f = vmap(polar_pt_vmap)(jnp.arange(v.size)).sum(axis=0)

    # we sum distance vectors, so they need to be in xyz for that to work
    # but then need to convert vectors back to rpz
    if kernel.ndim == 3:
        f = xyz2rpz_vec(f, phi=eval_data["phi"])

    return f


def singular_integral(
    eval_data,
    source_data,
    kernel,
    interpolator,
    loop=False,
):
    """Evaluate a singular integral transform on a surface.

    eg f(θ, ζ) = ∫ ∫ K(θ, ζ, θ', ζ') g(θ', ζ') dθ' dζ'

    Where K(θ, ζ, θ', ζ') is the (singular) kernel and g(θ', ζ') is the metric on the
    surface. See eq. 3.7 in [1]_, but we have absorbed the density σ into K

    Uses method by Malhotra et. al. [1]_ [2]_

    Parameters
    ----------
    eval_data : dict
        Dictionary of data at evaluation points (eval_grid passed to interpolator).
        Keys should be those required by kernel as kernel.keys. Vector data should be
        in rpz basis.
    source_data : dict
        Dictionary of data at source points (source_grid passed to interpolator). Keys
        should be those required by kernel as kernel.keys. Vector data should be in
        rpz basis.
    kernel : str or callable
        Kernel function to evaluate. If str, one of the following:
            '1_over_r' : 1 / |𝐫 − 𝐫'|
            'nr_over_r3' : 𝐧'⋅(𝐫 − 𝐫') / |𝐫 − 𝐫'|³
            'biot_savart' : μ₀/4π 𝐊'×(𝐫 − 𝐫') / |𝐫 − 𝐫'|³
            'biot_savart_A' : μ₀/4π 𝐊' / |𝐫 − 𝐫'|
        If callable, should take 3 arguments:
            eval_data : dict of data at evaluation points (primed)
            source_data : dict of data at source points (unprimed)
            diag : boolean, whether to evaluate full cross interactions or just diagonal
        If a callable, should also have the attributes ``ndim`` and ``keys`` defined.
        ``ndim`` is an integer representing the dimensionality of the output function f,
        1 if f is scalar, 3 if f is a vector, etc.
        ``keys`` is a list of strings of what data is required to evaluate the kernel.
        The kernel will be called with dictionaries containing this data at source and
        evaluation points.
        If vector valued, the input to the kernel function will be in rpz and output
        should be in xyz.
    interpolator : callable
        Function to interpolate from rectangular source grid to polar
        source grid around each singular point. See ``FFTInterpolator`` or
        ``DFTInterpolator``
    loop : bool
        If True, evaluate integral using loops, as opposed to vmap. Slower, but uses
        less memory.

    Returns
    -------
    f : ndarray, shape(eval_grid.num_nodes, kernel.ndim)
        Integral transform evaluated at eval_grid. Vectors are in rpz basis.

    References
    ----------
    .. [1] Malhotra, Dhairya, et al. "Efficient high-order singular quadrature schemes
       in magnetic fusion." Plasma Physics and Controlled Fusion 62.2 (2019): 024004.
    .. [2] Malhotra, Dhairya, et al. "Taylor states in stellarators: A fast high-order
       boundary integral solver." Journal of Computational Physics 397 (2019): 108791.

    """
    # sanitize inputs, we need everything as jax arrays so they can be indexed
    # properly in the loops
    source_data = {key: jnp.asarray(val) for key, val in source_data.items()}
    eval_data = {key: jnp.asarray(val) for key, val in eval_data.items()}

    if isinstance(kernel, str):
        kernel = kernels[kernel]

    s, q = interpolator.s, interpolator.q
    eval_grid, source_grid = interpolator._eval_grid, interpolator._source_grid

    out2 = _singular_part(
        eval_data, eval_grid, source_data, source_grid, s, q, kernel, interpolator, loop
    )
    out1 = _nonsingular_part(
        eval_data, eval_grid, source_data, source_grid, s, kernel, loop
    )
    return out1 + out2


def _kernel_nr_over_r3(eval_data, source_data, diag=False):
    # n * r / |r|^3
    source_x = jnp.atleast_2d(
        rpz2xyz(jnp.array([source_data["R"], source_data["phi"], source_data["Z"]]).T)
    )
    eval_x = jnp.atleast_2d(
        rpz2xyz(jnp.array([eval_data["R"], eval_data["phi"], eval_data["Z"]]).T)
    )
    if diag:
        dx = eval_x - source_x
    else:
        dx = eval_x[:, None] - source_x[None]
    n = rpz2xyz_vec(source_data["e^rho"], phi=source_data["phi"])
    n = n / jnp.linalg.norm(n, axis=-1)[:, None]
    r = safenorm(dx, axis=-1)
    return safediv(jnp.sum(n * dx, axis=-1), r**3)


_kernel_nr_over_r3.ndim = 1
_kernel_nr_over_r3.keys = ["R", "phi", "Z", "e^rho"]


def _kernel_1_over_r(eval_data, source_data, diag=False):
    # 1/ |r|
    source_x = jnp.atleast_2d(
        rpz2xyz(jnp.array([source_data["R"], source_data["phi"], source_data["Z"]]).T)
    )
    eval_x = jnp.atleast_2d(
        rpz2xyz(jnp.array([eval_data["R"], eval_data["phi"], eval_data["Z"]]).T)
    )
    if diag:
        dx = eval_x - source_x
    else:
        dx = eval_x[:, None] - source_x[None]
    r = safenorm(dx, axis=-1)
    return safediv(1, r)


_kernel_1_over_r.ndim = 1
_kernel_1_over_r.keys = ["R", "phi", "Z"]


def _kernel_biot_savart(eval_data, source_data, diag=False):
    # K x r / |r|^3
    source_x = jnp.atleast_2d(
        rpz2xyz(jnp.array([source_data["R"], source_data["phi"], source_data["Z"]]).T)
    )
    eval_x = jnp.atleast_2d(
        rpz2xyz(jnp.array([eval_data["R"], eval_data["phi"], eval_data["Z"]]).T)
    )
    if diag:
        dx = eval_x - source_x
    else:
        dx = eval_x[:, None] - source_x[None]
    K = rpz2xyz_vec(source_data["K_vc"], phi=source_data["phi"])
    num = jnp.cross(K, dx, axis=-1)
    r = safenorm(dx, axis=-1)
    if diag:
        r = r[:, None]
    else:
        r = r[:, :, None]
    return 1e-7 * safediv(num, r**3)


_kernel_biot_savart.ndim = 3
_kernel_biot_savart.keys = ["R", "phi", "Z", "K_vc"]


def _kernel_biot_savart_A(eval_data, source_data, diag=False):
    # K  / |r|
    source_x = jnp.atleast_2d(
        rpz2xyz(jnp.array([source_data["R"], source_data["phi"], source_data["Z"]]).T)
    )
    eval_x = jnp.atleast_2d(
        rpz2xyz(jnp.array([eval_data["R"], eval_data["phi"], eval_data["Z"]]).T)
    )
    if diag:
        dx = eval_x - source_x
    else:
        dx = eval_x[:, None] - source_x[None]
    K = rpz2xyz_vec(source_data["K_vc"], phi=source_data["phi"])
    r = safenorm(dx, axis=-1)
    if diag:
        r = r[:, None]
    else:
        r = r[:, :, None]
    return 1e-7 * safediv(K, r)


_kernel_biot_savart_A.ndim = 3
_kernel_biot_savart_A.keys = ["R", "phi", "Z", "K_vc"]


kernels = {
    "1_over_r": _kernel_1_over_r,
    "nr_over_r3": _kernel_nr_over_r3,
    "biot_savart": _kernel_biot_savart,
    "biot_savart_A": _kernel_biot_savart_A,
}


def virtual_casing_biot_savart(eval_data, source_data, interpolator, loop=True):
    """Evaluate magnetic field on surface due to sheet current on surface.

    The magnetic field due to the plasma current can be written as a Biot-Savart
    integral over the plasma volume:

    𝐁ᵥ(𝐫) = μ₀/4π ∫ 𝐉(𝐫') × (𝐫 − 𝐫')/|𝐫 − 𝐫'|³ d³𝐫'

    Where 𝐉 is the plasma current density, 𝐫 is a point on the plasma surface, and 𝐫' is
    a point in the plasma volume.

    This 3D integral can be converted to a 2D integral over the plasma boundary using
    the virtual casing principle [1]_

    𝐁ᵥ(𝐫) = μ₀/4π ∫ (𝐧' ⋅ 𝐁(𝐫')) (𝐫 − 𝐫')/|𝐫 − 𝐫'|³ d²𝐫'
            + μ₀/4π ∫ (𝐧' × 𝐁(𝐫') × (𝐫 − 𝐫')/ |𝐫 − 𝐫'|³ d²𝐫'
            + 𝐁(𝐫)/2

    Where 𝐁 is the total field on the surface and 𝐧' is the outward surface normal.
    Because the total field is tangent, the first term in the integrand is zero leaving

    𝐁ᵥ(𝐫) = μ₀/4π ∫ K_vc(𝐫') × (𝐫 − 𝐫')/ |𝐫 − 𝐫'|³ d²𝐫' + 𝐁(𝐫)/2

    Where we have defined the virtual casing sheet current K_vc = 𝐧' × 𝐁(𝐫')

    Parameters
    ----------
    eval_data : dict
        Dictionary of data at evaluation points (eval_grid passed to interpolator).
        Keys should be those required by kernel as kernel.keys. Vector data should be
        in rpz basis.
    source_data : dict
        Dictionary of data at source points (source_grid passed to interpolator). Keys
        should be those required by kernel as kernel.keys. Vector data should be in
        rpz basis.
    interpolator : callable
        Function to interpolate from rectangular source grid to polar
        source grid around each singular point. See ``FFTInterpolator`` or
        ``DFTInterpolator``
    loop : bool
        If True, evaluate integral using loops, as opposed to vmap. Slower, but uses
        less memory.

    Returns
    -------
    f : ndarray, shape(eval_grid.num_nodes, kernel.ndim)
        Integral transform evaluated at eval_grid. Vectors are in rpz basis.

    References
    ----------
    .. [1] Hanson, James D. "The virtual-casing principle and Helmholtz’s theorem."
       Plasma Physics and Controlled Fusion 57.11 (2015): 115006.

    """
    return singular_integral(
        eval_data,
        source_data,
        _kernel_biot_savart,
        interpolator,
        loop,
    )


def compute_B_plasma(eq, eval_grid, source_grid=None, normal_only=False):
    """Evaluate magnetic field on surface due to enclosed plasma currents.

    The magnetic field due to the plasma current can be written as a Biot-Savart
    integral over the plasma volume:

    𝐁ᵥ(𝐫) = μ₀/4π ∫ 𝐉(𝐫') × (𝐫 − 𝐫')/|𝐫 − 𝐫'|³ d³𝐫'

    Where 𝐉 is the plasma current density, 𝐫 is a point on the plasma surface, and 𝐫' is
    a point in the plasma volume.

    This 3D integral can be converted to a 2D integral over the plasma boundary using
    the virtual casing principle [1]_

    𝐁ᵥ(𝐫) = μ₀/4π ∫ (𝐧' ⋅ 𝐁(𝐫')) (𝐫 − 𝐫')/|𝐫 − 𝐫'|³ d²𝐫'
            + μ₀/4π ∫ (𝐧' × 𝐁(𝐫') × (𝐫 − 𝐫')/ |𝐫 − 𝐫'|³ d²𝐫'
            + 𝐁(𝐫)/2

    Where 𝐁 is the total field on the surface and 𝐧' is the outward surface normal.
    Because the total field is tangent, the first term in the integrand is zero leaving

    𝐁ᵥ(𝐫) = μ₀/4π ∫ K_vc(𝐫') × (𝐫 − 𝐫')/ |𝐫 − 𝐫'|³ d²𝐫' + 𝐁(𝐫)/2

    Where we have defined the virtual casing sheet current K_vc = 𝐧' × 𝐁(𝐫')

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that is the source of the plasma current.
    eval_grid : Grid
        Evaluation points for the magnetic field.
    source_grid : Grid, optional
        Source points for integral.
    normal_only : bool
        If True, only compute and return the normal component of the plasma field 𝐁ᵥ⋅𝐧

    Returns
    -------
    f : ndarray, shape(eval_grid.num_nodes, 3) or shape(eval_grid.num_nodes,)
        Magnetic field evaluated at eval_grid.
        If normal_only=False, vector B is in rpz basis.

    References
    ----------
    .. [1] Hanson, James D. "The virtual-casing principle and Helmholtz’s theorem."
       Plasma Physics and Controlled Fusion 57.11 (2015): 115006.

    """
    if source_grid is None:
        source_NFP = eq.NFP if eq.N > 0 else 64
        source_grid = LinearGrid(
            rho=np.array([1.0]),
            M=eq.M_grid,
            N=eq.N_grid,
            NFP=source_NFP,
            sym=False,
        )

    data_keys = ["K_vc", "B", "R", "phi", "Z", "e^rho", "n_rho", "|e_theta x e_zeta|"]
    eval_data = eq.compute(data_keys, grid=eval_grid)
    source_data = eq.compute(data_keys, grid=source_grid)

    k = min(source_grid.num_theta, source_grid.num_zeta * source_grid.NFP)
    s = k - 1
    q = k // 2 + int(np.sqrt(k))
    try:
        interpolator = FFTInterpolator(eval_grid, source_grid, s, q)
    except AssertionError as e:
        print(
            f"Unable to create FFTInterpolator, got error {e},"
            "falling back to DFT method which is much slower"
        )
        interpolator = DFTInterpolator(eval_grid, source_grid, s, q)
    if hasattr(eq.surface, "Phi_mn"):
        source_data["K_vc"] += eq.surface.compute("K", grid=source_grid)["K"]
    Bplasma = virtual_casing_biot_savart(eval_data, source_data, interpolator)
    # need extra factor of B/2 bc we're evaluating on plasma surface
    Bplasma = Bplasma + eval_data["B"] / 2
    if normal_only:
        Bplasma = jnp.sum(Bplasma * eval_data["n_rho"], axis=1)
    return Bplasma



"""Utility functions for the ``batched_vectorize`` function."""

import functools
from functools import partial
from typing import Callable, Optional

from jax._src.api import (
    _check_input_dtype_jacfwd,
    _check_input_dtype_jacrev,
    _check_output_dtype_jacfwd,
    _check_output_dtype_jacrev,
    _jacfwd_unravel,
    _jacrev_unravel,
    _jvp,
    _std_basis,
    _vjp,
)
from jax._src.api_util import _ensure_index, argnums_partial, check_callable
from jax._src.tree_util import tree_map, tree_structure, tree_transpose
from jax._src.util import wraps

from desc.backend import jax, jnp

if jax.__version_info__ >= (0, 4, 16):
    from jax.extend import linear_util as lu
else:
    from jax import linear_util as lu

from jax._src.numpy.vectorize import (
    _apply_excluded,
    _check_output_dims,
    _parse_gufunc_signature,
    _parse_input_dimensions,
)

# The following section of this code is derived from the NetKet project
# https://github.com/netket/netket/blob/9881c9fb217a2ac4dc9274a054bf6e6a2993c519/
# netket/jax/_chunk_utils.py
#
# The original copyright notice is as follows
# Copyright 2021 The NetKet Authors - All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");


def _treeify(f):
    def _f(x, *args, **kwargs):
        return jax.tree_util.tree_map(lambda y: f(y, *args, **kwargs), x)

    return _f


@_treeify
def _unchunk(x):
    return x.reshape((-1,) + x.shape[2:])


@_treeify
def _chunk(x, chunk_size=None):
    # chunk_size=None -> add just a dummy chunk dimension,
    #  same as np.expand_dims(x, 0)
    if x.ndim == 0:
        raise ValueError("x cannot be chunked as it has 0 dimensions.")
    n = x.shape[0]
    if chunk_size is None:
        chunk_size = n

    n_chunks, residual = divmod(n, chunk_size)
    if residual != 0:
        raise ValueError(
            "The first dimension of x must be divisible by chunk_size."
            + f"\n        Got x.shape={x.shape} but chunk_size={chunk_size}."
        )
    return x.reshape((n_chunks, chunk_size) + x.shape[1:])


####

# The following section of this code is derived from the NetKet project
# https://github.com/netket/netket/blob/9881c9fb217a2ac4dc9274a054bf6e6a2993c519/
# netket/jax/_scanmap.py


def scan_append(f, x):
    """Evaluate f element by element in x while appending the results.

    Parameters
    ----------
    f: a function that takes elements of the leading dimension of x
    x: a pytree where each leaf array has the same leading dimension

    Returns
    -------
    a (pytree of) array(s) with leading dimension same as x,
    containing the evaluation of f at each element in x
    """
    carry_init = True

    def f_(carry, x):
        return False, f(x)

    _, res_append = jax.lax.scan(f_, carry_init, x, unroll=1)
    return res_append


def _scanmap(fun, scan_fun, argnums=0):
    """A helper function to wrap f with a scan_fun."""

    def f_(*args, **kwargs):
        f = lu.wrap_init(fun, kwargs)
        f_partial, dyn_args = jax.api_util.argnums_partial(
            f, argnums, args, require_static_args_hashable=False
        )
        return scan_fun(lambda x: f_partial.call_wrapped(*x), dyn_args)

    return f_


# The following section of this code is derived from the NetKet project
# https://github.com/netket/netket/blob/9881c9fb217a2ac4dc9274a054bf6e6a2993c519/
# netket/jax/_vmap_chunked.py


def _eval_fun_in_chunks(vmapped_fun, chunk_size, argnums, *args, **kwargs):
    n_elements = jax.tree_util.tree_leaves(args[argnums[0]])[0].shape[0]
    n_chunks, n_rest = divmod(n_elements, chunk_size)

    if n_chunks == 0 or chunk_size >= n_elements:
        y = vmapped_fun(*args, **kwargs)
    else:
        # split inputs
        def _get_chunks(x):
            x_chunks = jax.tree_util.tree_map(
                lambda x_: x_[: n_elements - n_rest, ...], x
            )
            x_chunks = _chunk(x_chunks, chunk_size)
            return x_chunks

        def _get_rest(x):
            x_rest = jax.tree_util.tree_map(
                lambda x_: x_[n_elements - n_rest :, ...], x
            )
            return x_rest

        args_chunks = [
            _get_chunks(a) if i in argnums else a for i, a in enumerate(args)
        ]
        args_rest = [_get_rest(a) if i in argnums else a for i, a in enumerate(args)]

        y_chunks = _unchunk(
            _scanmap(vmapped_fun, scan_append, argnums)(*args_chunks, **kwargs)
        )

        if n_rest == 0:
            y = y_chunks
        else:
            y_rest = vmapped_fun(*args_rest, **kwargs)
            y = jax.tree_util.tree_map(
                lambda y1, y2: jnp.concatenate((y1, y2)), y_chunks, y_rest
            )
    return y


def _chunk_vmapped_function(
    vmapped_fun: Callable,
    chunk_size: Optional[int],
    argnums=0,
) -> Callable:
    """Takes a vmapped function and computes it in chunks."""
    if chunk_size is None:
        return vmapped_fun

    if isinstance(argnums, int):
        argnums = (argnums,)
    return functools.partial(_eval_fun_in_chunks, vmapped_fun, chunk_size, argnums)


def _parse_in_axes(in_axes):
    if isinstance(in_axes, int):
        in_axes = (in_axes,)

    if not set(in_axes).issubset((0, None)):
        raise NotImplementedError("Only in_axes 0/None are currently supported")

    argnums = tuple(
        map(lambda ix: ix[0], filter(lambda ix: ix[1] is not None, enumerate(in_axes)))
    )
    return in_axes, argnums


def vmap_chunked(
    f: Callable,
    in_axes=0,
    *,
    chunk_size: Optional[int],
) -> Callable:
    """Behaves like jax.vmap but uses scan to chunk the computations in smaller chunks.

    Parameters
    ----------
    f: The function to be vectorised.
    in_axes: The axes that should be scanned along. Only supports `0` or `None`
    chunk_size: The maximum size of the chunks to be used. If it is `None`,
        chunking is disabled


    Returns
    -------
    f: A vectorised and chunked function
    """
    in_axes, argnums = _parse_in_axes(in_axes)
    vmapped_fun = jax.vmap(f, in_axes=in_axes)
    return _chunk_vmapped_function(vmapped_fun, chunk_size, argnums)


def batched_vectorize(pyfunc, *, excluded=frozenset(), signature=None, chunk_size=None):
    """Define a vectorized function with broadcasting and batching.

    :func:`vectorize` is a convenience wrapper for defining vectorized
    functions with broadcasting, in the style of NumPy's
    `generalized universal functions
    <https://numpy.org/doc/stable/reference/c-api/generalized-ufuncs.html>`_.
    It allows for defining functions that are automatically repeated across
    any leading dimensions, without the implementation of the function needing to
    be concerned about how to handle higher dimensional inputs.

    :func:`jax.numpy.vectorize` has the same interface as
    :class:`numpy.vectorize`, but it is syntactic sugar for an auto-batching
    transformation (:func:`vmap`) rather than a Python loop. This should be
    considerably more efficient, but the implementation must be written in terms
    of functions that act on JAX arrays.

    Parameters
    ----------
    pyfunc: callable,function to vectorize.
    excluded: optional set of integers representing positional arguments for
    which the function will not be vectorized. These will be passed directly
    to ``pyfunc`` unmodified.
    signature: optional generalized universal function signature, e.g.,
    ``(m,n),(n)->(m)`` for vectorized matrix-vector multiplication. If
    provided, ``pyfunc`` will be called with (and expected to return) arrays
    with shapes given by the size of corresponding core dimensions. By
    default, pyfunc is assumed to take scalars arrays as input and output.
    chunk_size: the size of the batches to pass to vmap. If None, defaults to
    the largest possible chunk_size (like the default behavior of ``vectorize11)

    Returns
    -------
    Batch-vectorized version of the given function.

    """
    if any(not isinstance(exclude, (str, int)) for exclude in excluded):
        raise TypeError(
            "jax.numpy.vectorize can only exclude integer or string arguments, "
            "but excluded={!r}".format(excluded)
        )
    if any(isinstance(e, int) and e < 0 for e in excluded):
        raise ValueError(f"excluded={excluded!r} contains negative numbers")

    @functools.wraps(pyfunc)
    def wrapped(*args, **kwargs):
        error_context = (
            "on vectorized function with excluded={!r} and "
            "signature={!r}".format(excluded, signature)
        )
        excluded_func, args, kwargs = _apply_excluded(pyfunc, excluded, args, kwargs)

        if signature is not None:
            input_core_dims, output_core_dims = _parse_gufunc_signature(signature)
        else:
            input_core_dims = [()] * len(args)
            output_core_dims = None

        none_args = {i for i, arg in enumerate(args) if arg is None}
        if any(none_args):
            if any(input_core_dims[i] != () for i in none_args):
                raise ValueError(
                    f"Cannot pass None at locations {none_args} with {signature=}"
                )
            excluded_func, args, _ = _apply_excluded(excluded_func, none_args, args, {})
            input_core_dims = [
                dim for i, dim in enumerate(input_core_dims) if i not in none_args
            ]

        args = tuple(map(jnp.asarray, args))

        broadcast_shape, dim_sizes = _parse_input_dimensions(
            args, input_core_dims, error_context
        )

        checked_func = _check_output_dims(
            excluded_func, dim_sizes, output_core_dims, error_context
        )

        # Rather than broadcasting all arguments to full broadcast shapes, prefer
        # expanding dimensions using vmap. By pushing broadcasting
        # into vmap, we can make use of more efficient batching rules for
        # primitives where only some arguments are batched (e.g., for
        # lax_linalg.triangular_solve), and avoid instantiating large broadcasted
        # arrays.

        squeezed_args = []
        rev_filled_shapes = []

        for arg, core_dims in zip(args, input_core_dims):
            noncore_shape = arg.shape[: arg.ndim - len(core_dims)]

            pad_ndim = len(broadcast_shape) - len(noncore_shape)
            filled_shape = pad_ndim * (1,) + noncore_shape
            rev_filled_shapes.append(filled_shape[::-1])

            squeeze_indices = tuple(
                i for i, size in enumerate(noncore_shape) if size == 1
            )
            squeezed_arg = jnp.squeeze(arg, axis=squeeze_indices)
            squeezed_args.append(squeezed_arg)

        vectorized_func = checked_func
        dims_to_expand = []
        for negdim, axis_sizes in enumerate(zip(*rev_filled_shapes)):
            in_axes = tuple(None if size == 1 else 0 for size in axis_sizes)
            if all(axis is None for axis in in_axes):
                dims_to_expand.append(len(broadcast_shape) - 1 - negdim)
            else:
                # change the vmap here to chunked_vmap
                vectorized_func = vmap_chunked(
                    vectorized_func, in_axes, chunk_size=chunk_size
                )
        result = vectorized_func(*squeezed_args)

        if not dims_to_expand:
            return result
        elif isinstance(result, tuple):
            return tuple(jnp.expand_dims(r, axis=dims_to_expand) for r in result)
        else:
            return jnp.expand_dims(result, axis=dims_to_expand)

    return wrapped


# The following section of this code is derived from JAX
# https://github.com/jax-ml/jax/blob/ff0a98a2aef958df156ca149809cf532efbbcaf4/
# jax/_src/api.py
#
# The original copyright notice is as follows
# Copyright 2018 The JAX Authors.
# Licensed under the Apache License, Version 2.0 (the "License");


def jacfwd_chunked(
    fun,
    argnums=0,
    has_aux=False,
    holomorphic=False,
    *,
    chunk_size=None,
):
    """Jacobian of ``fun`` evaluated column-by-column using forward-mode AD.

    Parameters
    ----------
    fun: callable
        Function whose Jacobian is to be computed.
    argnums: Optional, integer or sequence of integers.
        Specifies which positional argument(s) to differentiate with respect to
        (default ``0``).
    has_aux: Optional, bool.
        Indicates whether ``fun`` returns a pair where the first element is considered
        the output of the mathematical function to be differentiated and the second
        element is auxiliary data. Default False.
    holomorphic: Optional, bool.
        Indicates whether ``fun`` is promised to be holomorphic. Default False.
    chunk_size: int
        The size of the batches to pass to vmap. If None, defaults to the largest
        possible chunk_size.

    Returns
    -------
    jac: callable
        A function with the same arguments as ``fun``, that evaluates the Jacobian of
        ``fun`` using forward-mode automatic differentiation. If ``has_aux`` is True
        then a pair of (jacobian, auxiliary_data) is returned.

    """
    check_callable(fun)
    argnums = _ensure_index(argnums)

    docstr = (
        "Jacobian of {fun} with respect to positional argument(s) "
        "{argnums}. Takes the same arguments as {fun} but returns the "
        "jacobian of the output with respect to the arguments at "
        "positions {argnums}."
    )

    @wraps(fun, docstr=docstr, argnums=argnums)
    def jacfun(*args, **kwargs):
        f = lu.wrap_init(fun, kwargs)
        f_partial, dyn_args = argnums_partial(
            f, argnums, args, require_static_args_hashable=False
        )
        tree_map(partial(_check_input_dtype_jacfwd, holomorphic), dyn_args)
        if not has_aux:
            pushfwd: Callable = partial(_jvp, f_partial, dyn_args)
            y, jac = vmap_chunked(pushfwd, chunk_size=chunk_size)(_std_basis(dyn_args))
            y = tree_map(lambda x: x[0], y)
            jac = tree_map(lambda x: jnp.moveaxis(x, 0, -1), jac)
        else:
            pushfwd: Callable = partial(_jvp, f_partial, dyn_args, has_aux=True)
            y, jac, aux = vmap_chunked(pushfwd, chunk_size=chunk_size)(
                _std_basis(dyn_args)
            )
            y = tree_map(lambda x: x[0], y)
            jac = tree_map(lambda x: jnp.moveaxis(x, 0, -1), jac)
            aux = tree_map(lambda x: x[0], aux)
        tree_map(partial(_check_output_dtype_jacfwd, holomorphic), y)
        example_args = dyn_args[0] if isinstance(argnums, int) else dyn_args
        jac_tree = tree_map(partial(_jacfwd_unravel, example_args), y, jac)
        if not has_aux:
            return jac_tree
        else:
            return jac_tree, aux

    return jacfun


def jacrev_chunked(
    fun,
    argnums=0,
    has_aux=False,
    holomorphic=False,
    allow_int=False,
    *,
    chunk_size=None,
):
    """Jacobian of ``fun`` evaluated row-by-row using reverse-mode AD.

    Parameters
    ----------
    fun: callable
        Function whose Jacobian is to be computed.
    argnums: Optional, integer or sequence of integers.
        Specifies which positional argument(s) to differentiate with respect to
        (default ``0``).
    has_aux: Optional, bool.
        Indicates whether ``fun`` returns a pair where the first element is considered
        the output of the mathematical function to be differentiated and the second
        element is auxiliary data. Default False.
    holomorphic: Optional, bool.
        Indicates whether ``fun`` is promised to be holomorphic. Default False.
    allow_int: Optional, bool.
        Whether to allow differentiating with respect to integer valued inputs. The
        gradient of an integer input will have a trivial vector-space dtype (float0).
        Default False.
    chunk_size: int
        The size of the batches to pass to vmap. If None, defaults to the largest
        possible chunk_size.

    Returns
    -------
    jac: callable
        A function with the same arguments as ``fun``, that evaluates the Jacobian of
        ``fun`` using reverse-mode automatic differentiation. If ``has_aux`` is True
        then a pair of (jacobian, auxiliary_data) is returned.

    """
    check_callable(fun)

    docstr = (
        "Jacobian of {fun} with respect to positional argument(s) "
        "{argnums}. Takes the same arguments as {fun} but returns the "
        "jacobian of the output with respect to the arguments at "
        "positions {argnums}."
    )

    @wraps(fun, docstr=docstr, argnums=argnums)
    def jacfun(*args, **kwargs):
        f = lu.wrap_init(fun, kwargs)
        f_partial, dyn_args = argnums_partial(
            f, argnums, args, require_static_args_hashable=False
        )
        tree_map(partial(_check_input_dtype_jacrev, holomorphic, allow_int), dyn_args)
        if not has_aux:
            y, pullback = _vjp(f_partial, *dyn_args)
        else:
            y, pullback, aux = _vjp(f_partial, *dyn_args, has_aux=True)
        tree_map(partial(_check_output_dtype_jacrev, holomorphic), y)
        jac = vmap_chunked(pullback, chunk_size=chunk_size)(_std_basis(y))
        jac = jac[0] if isinstance(argnums, int) else jac
        example_args = dyn_args[0] if isinstance(argnums, int) else dyn_args
        jac_tree = tree_map(partial(_jacrev_unravel, y), example_args, jac)
        jac_tree = tree_transpose(
            tree_structure(example_args), tree_structure(y), jac_tree
        )
        if not has_aux:
            return jac_tree
        else:
            return jac_tree, aux

    return jacfun



"""Tests for reading/writing/converting VMEC data."""

import numpy as np
import pytest
from netCDF4 import Dataset

import desc.examples
from desc.basis import DoubleFourierSeries, FourierZernikeBasis
from desc.equilibrium import EquilibriaFamily, Equilibrium
from desc.examples import get
from desc.grid import LinearGrid
from desc.input_reader import InputReader
from desc.io import load
from desc.vmec import VMECIO
from desc.vmec_utils import (
    fourier_to_zernike,
    ptolemy_identity_fwd,
    ptolemy_identity_rev,
    ptolemy_linear_transform,
    vmec_boundary_subspace,
    zernike_to_fourier,
)

from .utils import area_difference_desc


class TestVMECIO:
    """Tests VMECIO class."""

    @pytest.mark.unit
    def test_ptolemy_identity_fwd(self):
        """Tests forward implementation of Ptolemy's identity."""
        a0 = 3
        a1 = -1
        a2 = 1
        a3 = 2

        m_0 = np.array([0, 1, 1, 1, 0])
        n_0 = np.array([1, -1, 0, 1, 0])
        s = np.array([a0, a1, a3, 0, 0])
        c = np.array([0, a2, 0, a3, a0])

        # a0*sin(-z) + a1*sin(t+z) + a3*sin(t)                          # noqa: E800
        #    = -a0*sin(z) + a1*sin(t)*cos(z) + a1*cos(t)*sin(z) + a3*sin(t)
        # a0 + a2*cos(t+z) + a3*cos(t-z)                                # noqa: E800
        #    = a0 + (a2+a3)*cos(t)*cos(z) + (a3-a2)*sin(t)*sin(z)

        m_1_correct = np.array([-1, 0, 1, -1, 0, 1, -1, 0, 1])
        n_1_correct = np.array([-1, -1, -1, 0, 0, 0, 1, 1, 1])
        x_correct = np.array([[a3 - a2, -a0, a1, a3, a0, 0, a1, 0, a2 + a3]])

        m_1, n_1, x = ptolemy_identity_fwd(m_0, n_0, s, c)

        np.testing.assert_allclose(m_1, m_1_correct, atol=1e-8)
        np.testing.assert_allclose(n_1, n_1_correct, atol=1e-8)
        np.testing.assert_allclose(x, x_correct, atol=1e-8)

    @pytest.mark.unit
    def test_ptolemy_identity_fwd_sin_series(self):
        """Tests forward implementation of Ptolemy's identity for sin only."""
        a1 = -1

        m_0 = np.array([1, 1])
        n_0 = np.array([-1, 1])
        s = np.array([[a1, 0]])
        c = np.array([[0, 0]])

        m_1, n_1, x = ptolemy_identity_fwd(m_0, n_0, s, c)

        print(m_1)
        print(n_1)
        print(x)

        #  a1*sin(t+z)  # noqa: E800
        # = a1*sin(t)*cos(z) + a1*cos(t)*sin(z) # noqa: E800

        m_1_correct = np.array([-1, 1, -1, 1])
        n_1_correct = np.array([-1, -1, 1, 1])
        x_correct = np.array([[0, a1, a1, 0]])

        np.testing.assert_allclose(m_1, m_1_correct, atol=1e-8)
        np.testing.assert_allclose(n_1, n_1_correct, atol=1e-8)
        np.testing.assert_allclose(x, x_correct, atol=1e-8)

    @pytest.mark.unit
    def test_ptolemy_identity_rev(self):
        """Tests reverse implementation of Ptolemy's identity."""
        a0 = 3
        a1 = -1
        a2 = 1
        a3 = 2

        m_1 = np.array([-1, -1, -1, 0, 0, 0, 1, 1, 1])
        n_1 = np.array([-1, 0, 1, -1, 0, 1, -1, 0, 1])
        x = np.array([[a3 - a2, a3, a1, -a0, a0, 0, a1, 0, a2 + a3]])

        # -a0*sin(z) + a1*sin(t)*cos(z) + a1*cos(t)*sin(z) + a3*sin(t)   # noqa: E800
        #   = a0*sin(-z) + a1*sin(t+z) + a3*sin(t)
        # a0 + (a2+a3)*cos(t)*cos(z) + (a3-a2)*sin(t)*sin(z)             # noqa: E800
        #    = a0 + a2*cos(t+z) + a3*cos(t-z)

        m_0_correct = np.array([0, 0, 1, 1, 1])
        n_0_correct = np.array([0, 1, -1, 0, 1])
        s_correct = np.array([[0, a0, a1, a3, 0]])
        c_correct = np.array([[a0, 0, a2, 0, a3]])

        m_0, n_0, s, c = ptolemy_identity_rev(m_1, n_1, x)

        np.testing.assert_allclose(m_0, m_0_correct, atol=1e-8)
        np.testing.assert_allclose(n_0, n_0_correct, atol=1e-8)
        np.testing.assert_allclose(s, s_correct, atol=1e-8)
        np.testing.assert_allclose(c, c_correct, atol=1e-8)

    @pytest.mark.unit
    def test_ptolemy_identity_rev_sin_sym(self):
        """Tests reverse implementation of Ptolemy's identity for sin series."""
        a1 = -1

        m_1 = np.array([-1, 1])
        n_1 = np.array([1, -1])
        x = np.array([[a1, a1]])

        # a1*sin(t)*cos(z) + a1*cos(t)*sin(z)    # noqa: E800
        #   = a1*sin(t+z)

        m_0_correct = np.array([0, 1, 1])
        n_0_correct = np.array([0, -1, 1])
        s_correct = np.array([[0, a1, 0]])
        c_correct = np.array([[0, 0, 0]])

        m_0, n_0, s, c = ptolemy_identity_rev(m_1, n_1, x)

        np.testing.assert_allclose(m_0, m_0_correct, atol=1e-8)
        np.testing.assert_allclose(n_0, n_0_correct, atol=1e-8)
        np.testing.assert_allclose(s, s_correct, atol=1e-8)
        np.testing.assert_allclose(c, c_correct, atol=1e-8)

    @pytest.mark.unit
    def test_ptolemy_identities_inverse(self):
        """Tests that forward and reverse Ptolemy's identities are inverses."""
        basis = DoubleFourierSeries(4, 3, sym=False)
        modes = basis.modes
        x_correct = np.random.rand(basis.num_modes)

        m1, n1, s, c = ptolemy_identity_rev(modes[:, 1], modes[:, 2], x_correct)
        m0, n0, x = ptolemy_identity_fwd(m1, n1, s, c)

        np.testing.assert_allclose(m0, modes[:, 1])
        np.testing.assert_allclose(n0, modes[:, 2])
        np.testing.assert_allclose(x, np.atleast_2d(x_correct))

    @pytest.mark.unit
    def test_ptolemy_linear_transform(self):
        """Tests Ptolemy basis linear transformation utility function."""
        basis = DoubleFourierSeries(M=4, N=3, sym=False)
        matrix, modes, idx = ptolemy_linear_transform(
            basis.modes, helicity=(1, 1), NFP=1
        )

        x_correct = np.random.rand(basis.num_modes)
        x_transformed = matrix @ x_correct

        x_sin = np.insert(x_transformed[1::2], 0, 0)
        x_cos = x_transformed[::2]

        m, n, s, c = ptolemy_identity_rev(
            basis.modes[:, 1], basis.modes[:, 2], x_correct
        )
        np.testing.assert_allclose(modes[::2, 1], m)
        np.testing.assert_allclose(modes[::2, 2], n)
        np.testing.assert_allclose(np.atleast_2d(x_sin), s)
        np.testing.assert_allclose(np.atleast_2d(x_cos), c)

        _, _, x_original = ptolemy_identity_fwd(
            modes[::2, 1], modes[::2, 2], x_sin, x_cos
        )
        np.testing.assert_allclose(x_original, np.atleast_2d(x_correct))

        sym_modes = np.array(
            [
                [1, 0, 0],
                [-1, 1, 1],
                [1, 1, 1],
                [-1, 2, 2],
                [1, 2, 2],
                [-1, 3, 3],
                [1, 3, 3],
            ]
        )
        np.testing.assert_allclose(np.delete(modes, idx, axis=0), sym_modes)

    @pytest.mark.unit
    def test_fourier_to_zernike(self):
        """Test conversion from radial-Fourier series to Fourier-Zernike polynomials."""
        M = 1
        N = 1

        a0 = 3
        a1 = -1
        a2 = 1
        a3 = 2

        surfs = 16
        rho = np.sqrt(np.linspace(0, 1, surfs))

        m = np.array([-1, -1, -1, 0, 0, 0, 1, 1, 1])
        n = np.array([-1, 0, 1, -1, 0, 1, -1, 0, 1])
        x = np.array([a3 - a2, a3, a1, -a0, a0, 0, a1, 0, a2 + a3])

        # x * rho^|m|
        x_mn = np.power(np.atleast_2d(rho).T, np.atleast_2d(np.abs(m))) * np.atleast_2d(
            x
        )
        basis = FourierZernikeBasis(L=M, M=M, N=N, spectral_indexing="ansi")
        x_lmn = fourier_to_zernike(m, n, x_mn, basis)

        x_lmn_correct = np.zeros((basis.num_modes,))
        for k in range(basis.num_modes):
            idx = np.where((basis.modes == [np.abs(m[k]), m[k], n[k]]).all(axis=1))[0]
            x_lmn_correct[idx] = x[k]

        np.testing.assert_allclose(x_lmn, x_lmn_correct, atol=1e-8)

    @pytest.mark.unit
    def test_zernike_to_fourier(self):
        """Test conversion from Fourier-Zernike polynomials to radial-Fourier series."""
        M = 1
        N = 1

        a0 = 3
        a1 = -1
        a2 = 1
        a3 = 2

        surfs = 16
        rho = np.sqrt(np.linspace(0, 1, surfs))

        m_correct = np.array([-1, -1, -1, 0, 0, 0, 1, 1, 1])
        n_correct = np.array([-1, 0, 1, -1, 0, 1, -1, 0, 1])
        x = np.array([a3 - a2, a3, a1, -a0, a0, 0, a1, 0, a2 + a3])

        # x * rho^|m|
        x_mn_correct = np.power(
            np.atleast_2d(rho).T, np.atleast_2d(np.abs(m_correct))
        ) * np.atleast_2d(x)
        basis = FourierZernikeBasis(L=M, M=M, N=N, spectral_indexing="ansi")

        x_lmn = np.zeros((basis.num_modes,))
        for k in range(basis.num_modes):
            idx = np.where(
                (basis.modes == [np.abs(m_correct[k]), m_correct[k], n_correct[k]]).all(
                    axis=1
                )
            )[0]
            x_lmn[idx] = x[k]

        m, n, x_mn = zernike_to_fourier(x_lmn, basis, rho)

        np.testing.assert_allclose(m, m_correct, atol=1e-8)
        np.testing.assert_allclose(n, n_correct, atol=1e-8)
        np.testing.assert_allclose(x_mn, x_mn_correct, atol=1e-8)


@pytest.mark.unit
def test_vmec_load_profiles(TmpDir):
    """Tests that loading with iota or current profiles give same result."""
    input_path = "./tests/inputs/wout_SOLOVEV.nc"

    eq_iota = VMECIO.load(input_path, profile="iota")
    eq_current = VMECIO.load(input_path, profile="current")

    assert eq_iota.current is None
    assert eq_current.iota is None

    grid = LinearGrid(
        M=eq_iota.M_grid,
        N=eq_iota.N_grid,
        NFP=eq_iota.NFP,
        rho=np.linspace(0.5, 1.0, 21),
    )
    data_iota = eq_iota.compute(["iota", "current"], grid=grid)
    data_current = eq_current.compute(["iota", "current"], grid=grid)

    iota_iota = grid.compress(data_iota["iota"])
    iota_current = grid.compress(data_current["iota"])
    current_iota = grid.compress(data_iota["current"])
    current_current = grid.compress(data_current["current"])

    np.testing.assert_allclose(iota_iota, iota_current, rtol=2e-2)
    np.testing.assert_allclose(current_current, current_iota, rtol=2e-2)


@pytest.mark.slow
@pytest.mark.unit
def test_load_then_save(TmpDir):
    """Tests if loading and then saving gives the original result."""
    input_path = "./tests/inputs/wout_SOLOVEV.nc"
    output_path = str(TmpDir.join("DESC_SOLOVEV.nc"))

    eq = VMECIO.load(input_path, profile="iota")
    VMECIO.save(eq, output_path)

    file1 = Dataset(input_path, mode="r")
    file2 = Dataset(output_path, mode="r")

    rmnc1 = file1.variables["rmnc"][:]
    rmnc2 = file2.variables["rmnc"][:]
    zmns1 = file1.variables["zmns"][:]
    zmns2 = file2.variables["zmns"][:]
    lmns1 = file1.variables["lmns"][:]
    lmns2 = file2.variables["lmns"][:]

    np.testing.assert_allclose(rmnc2, rmnc1, rtol=1e-3, atol=1e-3)
    np.testing.assert_allclose(zmns2, zmns1, rtol=1e-3, atol=1e-3)
    np.testing.assert_allclose(lmns2, lmns1, rtol=1e-3, atol=5e-2)

    file1.close()
    file2.close()


@pytest.mark.slow
@pytest.mark.unit
def test_axis_surf_after_load():
    """Tests if loading VMEC solution preserves axis and surface."""
    input_path = "./tests/inputs/wout_HELIOTRON.nc"

    eq = VMECIO.load(input_path, profile="iota")
    assert eq.is_nested()
    f = Dataset(input_path)

    surf1 = eq.surface
    surf2 = eq.get_surface_at(rho=1)
    axis1 = eq.axis
    axis2 = eq.get_axis()

    np.testing.assert_allclose(surf1.R_lmn, surf2.R_lmn, atol=1e-14)
    np.testing.assert_allclose(surf1.Z_lmn, surf2.Z_lmn, atol=1e-14)

    np.testing.assert_allclose(axis1.R_n, axis2.R_n, atol=1e-14)
    np.testing.assert_allclose(axis1.Z_n, axis2.Z_n, atol=1e-14)

    # surface
    rm, rn, Rb_cs, Rb_cc = ptolemy_identity_rev(
        surf1.R_basis.modes[:, 1],
        surf1.R_basis.modes[:, 2],
        np.where(surf1.R_basis.modes[:, 1] < 0, -surf1.R_lmn, surf1.R_lmn)[None, :],
    )

    zm, zn, Zb_cs, Zb_cc = ptolemy_identity_rev(
        surf1.Z_basis.modes[:, 1],
        surf1.Z_basis.modes[:, 2],
        np.where(surf1.Z_basis.modes[:, 1] < 0, -surf1.Z_lmn, surf1.Z_lmn)[None, :],
    )

    rmnc = f.variables["rmnc"][:].filled()
    zmns = f.variables["zmns"][:].filled()

    # axis
    rax_cc = f.variables["raxis_cc"][:].filled()
    zax_cs = -f.variables["zaxis_cs"][:].filled()

    np.testing.assert_allclose(rmnc[-1], Rb_cc.squeeze(), atol=1e-14)
    np.testing.assert_allclose(zmns[-1], Zb_cs.squeeze(), atol=1e-14)

    np.testing.assert_allclose(rax_cc, axis1.R_n, atol=1e-14)
    np.testing.assert_allclose(zax_cs[1:][::-1], axis1.Z_n, atol=1e-14)

    f.close()


@pytest.mark.unit
def test_vmec_save_kinetic(TmpDir):
    """Tests that saving an equilibrium with kinetic profiles runs without errors."""
    output_path = str(TmpDir.join("output.nc"))
    eq = Equilibrium(
        L=2,
        M=2,
        N=2,
        NFP=3,
        electron_density=np.array([[0, 1], [2, -1]]),
        electron_temperature=np.array([[0, 1], [2, -1]]),
        sym=True,
    )
    VMECIO.save(eq, output_path)


@pytest.mark.regression
@pytest.mark.slow
def test_vmec_save_1(VMEC_save):
    """Tests that saving in NetCDF format agrees with VMEC."""
    vmec, desc = VMEC_save
    # parameters
    assert vmec.variables["version_"][:] == desc.variables["version_"][:]
    assert vmec.variables["mgrid_mode"][:] == desc.variables["mgrid_mode"][:]
    assert np.all(
        np.char.compare_chararrays(
            vmec.variables["mgrid_file"][:],
            desc.variables["mgrid_file"][:],
            "==",
            False,
        )
    )
    assert vmec.variables["ier_flag"][:] == desc.variables["ier_flag"][:]
    assert (
        vmec.variables["lfreeb__logical__"][:] == desc.variables["lfreeb__logical__"][:]
    )
    assert (
        vmec.variables["lrecon__logical__"][:] == desc.variables["lrecon__logical__"][:]
    )
    assert vmec.variables["lrfp__logical__"][:] == desc.variables["lrfp__logical__"][:]
    assert (
        vmec.variables["lasym__logical__"][:] == desc.variables["lasym__logical__"][:]
    )
    assert vmec.variables["nfp"][:] == desc.variables["nfp"][:]
    assert vmec.variables["ns"][:] == desc.variables["ns"][:]
    assert vmec.variables["mpol"][:] == desc.variables["mpol"][:]
    assert vmec.variables["ntor"][:] == desc.variables["ntor"][:]
    assert vmec.variables["mnmax"][:] == desc.variables["mnmax"][:]
    np.testing.assert_allclose(vmec.variables["xm"][:], desc.variables["xm"][:])
    np.testing.assert_allclose(vmec.variables["xn"][:], desc.variables["xn"][:])
    assert vmec.variables["mnmax_nyq"][:] == desc.variables["mnmax_nyq"][:]
    np.testing.assert_allclose(vmec.variables["xm_nyq"][:], desc.variables["xm_nyq"][:])
    np.testing.assert_allclose(vmec.variables["xn_nyq"][:], desc.variables["xn_nyq"][:])
    assert vmec.variables["signgs"][:] == desc.variables["signgs"][:]
    assert vmec.variables["gamma"][:] == desc.variables["gamma"][:]
    assert vmec.variables["nextcur"][:] == desc.variables["nextcur"][:]
    assert np.all(
        np.char.compare_chararrays(
            vmec.variables["pmass_type"][:],
            desc.variables["pmass_type"][:],
            "==",
            False,
        )
    )
    assert np.all(
        np.char.compare_chararrays(
            vmec.variables["piota_type"][:],
            desc.variables["piota_type"][:],
            "==",
            False,
        )
    )
    assert np.all(
        np.char.compare_chararrays(
            vmec.variables["pcurr_type"][:],
            desc.variables["pcurr_type"][:],
            "==",
            False,
        )
    )
    np.testing.assert_allclose(
        vmec.variables["am"][:], desc.variables["am"][:], atol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["ai"][:], desc.variables["ai"][:], atol=1e-8
    )
    np.testing.assert_allclose(
        vmec.variables["ac"][:], desc.variables["ac"][:], atol=1e-8
    )
    np.testing.assert_allclose(
        vmec.variables["presf"][:], desc.variables["presf"][:], atol=2e-2
    )
    np.testing.assert_allclose(vmec.variables["pres"][:], desc.variables["pres"][:])
    np.testing.assert_allclose(vmec.variables["mass"][:], desc.variables["mass"][:])
    np.testing.assert_allclose(vmec.variables["iotaf"][:], desc.variables["iotaf"][:])
    np.testing.assert_allclose(
        vmec.variables["q_factor"][:], desc.variables["q_factor"][:]
    )
    np.testing.assert_allclose(vmec.variables["iotas"][:], desc.variables["iotas"][:])
    np.testing.assert_allclose(vmec.variables["phi"][:], desc.variables["phi"][:])
    np.testing.assert_allclose(vmec.variables["phipf"][:], desc.variables["phipf"][:])
    np.testing.assert_allclose(vmec.variables["phips"][:], desc.variables["phips"][:])
    np.testing.assert_allclose(
        vmec.variables["chi"][:], desc.variables["chi"][:], atol=2e-5
    )
    np.testing.assert_allclose(vmec.variables["chipf"][:], desc.variables["chipf"][:])
    np.testing.assert_allclose(
        vmec.variables["Rmajor_p"][:], desc.variables["Rmajor_p"][:]
    )
    np.testing.assert_allclose(
        vmec.variables["Aminor_p"][:], desc.variables["Aminor_p"][:]
    )
    np.testing.assert_allclose(vmec.variables["aspect"][:], desc.variables["aspect"][:])
    np.testing.assert_allclose(
        vmec.variables["volume_p"][:], desc.variables["volume_p"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["volavgB"][:], desc.variables["volavgB"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["betatotal"][:], desc.variables["betatotal"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["betapol"][:], desc.variables["betapol"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["betator"][:], desc.variables["betator"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["ctor"][:], desc.variables["ctor"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["rbtor"][:], desc.variables["rbtor"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["rbtor0"][:], desc.variables["rbtor0"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["b0"][:], desc.variables["b0"][:], rtol=5e-5
    )
    np.testing.assert_allclose(
        vmec.variables["buco"][20:100], desc.variables["buco"][20:100], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["bvco"][20:100], desc.variables["bvco"][20:100], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["vp"][20:100], desc.variables["vp"][20:100], rtol=1e-6
    )
    np.testing.assert_allclose(
        vmec.variables["bdotb"][20:100], desc.variables["bdotb"][20:100], rtol=1e-6
    )
    np.testing.assert_allclose(
        vmec.variables["jdotb"][20:100], desc.variables["jdotb"][20:100], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["jcuru"][20:100], desc.variables["jcuru"][20:100], rtol=1e-2
    )
    np.testing.assert_allclose(
        vmec.variables["jcurv"][20:100], desc.variables["jcurv"][20:100], rtol=3e-2
    )
    np.testing.assert_allclose(
        vmec.variables["DShear"][20:100], desc.variables["DShear"][20:100], rtol=1e-2
    )
    np.testing.assert_allclose(
        vmec.variables["DCurr"][20:100], desc.variables["DCurr"][20:100], rtol=1e-2
    )
    np.testing.assert_allclose(
        vmec.variables["DWell"][20:100], desc.variables["DWell"][20:100], rtol=1e-2
    )
    np.testing.assert_allclose(
        vmec.variables["DGeod"][20:100], desc.variables["DGeod"][20:100], atol=1e-9
    )
    np.testing.assert_allclose(
        vmec.variables["DMerc"][20:100], desc.variables["DMerc"][20:100], rtol=5e-2
    )
    np.testing.assert_allclose(
        vmec.variables["raxis_cc"][:], desc.variables["raxis_cc"][:], rtol=5e-5
    )
    np.testing.assert_allclose(
        vmec.variables["zaxis_cs"][:], desc.variables["zaxis_cs"][:], rtol=5e-5
    )
    np.testing.assert_allclose(
        vmec.variables["rmin_surf"][:], desc.variables["rmin_surf"][:], rtol=5e-3
    )
    np.testing.assert_allclose(
        vmec.variables["rmax_surf"][:], desc.variables["rmax_surf"][:], rtol=5e-3
    )
    np.testing.assert_allclose(
        vmec.variables["zmax_surf"][:], desc.variables["zmax_surf"][:], rtol=5e-3
    )
    np.testing.assert_allclose(
        vmec.variables["beta_vol"][:], desc.variables["beta_vol"][:], rtol=5e-5
    )
    np.testing.assert_allclose(
        vmec.variables["betaxis"][:], desc.variables["betaxis"][:], rtol=5e-5
    )


@pytest.mark.regression
@pytest.mark.slow
def test_vmec_save_2(VMEC_save):
    """Tests that saving in NetCDF format agrees with VMEC."""
    vmec, desc = VMEC_save

    # straight field-line grid to compare quantities in full volume
    grid = LinearGrid(L=15, M=6, N=0, NFP=desc.variables["nfp"][:])
    theta_vmec = VMECIO.compute_theta_coords(
        vmec.variables["lmns"][:],
        vmec.variables["xm"][:],
        vmec.variables["xn"][:],
        grid.nodes[:, 0],
        grid.nodes[:, 1],
        grid.nodes[:, 2],
    )
    theta_desc = VMECIO.compute_theta_coords(
        desc.variables["lmns"][:],
        desc.variables["xm"][:],
        desc.variables["xn"][:],
        grid.nodes[:, 0],
        grid.nodes[:, 1],
        grid.nodes[:, 2],
    )

    # R & Z
    R_vmec, Z_vmec = VMECIO.vmec_interpolate(
        vmec.variables["rmnc"][:],
        vmec.variables["zmns"][:],
        vmec.variables["xm"][:],
        vmec.variables["xn"][:],
        theta=theta_vmec,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=True,
    )
    R_desc, Z_desc = VMECIO.vmec_interpolate(
        desc.variables["rmnc"][:],
        desc.variables["zmns"][:],
        desc.variables["xm"][:],
        desc.variables["xn"][:],
        theta=theta_desc,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=True,
    )
    np.testing.assert_allclose(R_vmec, R_desc, rtol=1e-3)
    np.testing.assert_allclose(Z_vmec, Z_desc, rtol=1e-3)

    # |B|
    b_vmec = VMECIO.vmec_interpolate(
        vmec.variables["bmnc"][:],
        np.zeros_like(vmec.variables["bmnc"][:]),
        vmec.variables["xm_nyq"][:],
        vmec.variables["xn_nyq"][:],
        theta=theta_vmec,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    b_desc = VMECIO.vmec_interpolate(
        desc.variables["bmnc"][:],
        np.zeros_like(desc.variables["bmnc"][:]),
        desc.variables["xm_nyq"][:],
        desc.variables["xn_nyq"][:],
        theta=theta_desc,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    np.testing.assert_allclose(b_vmec, b_desc, rtol=1e-3)

    # B^zeta
    bsupv_vmec = VMECIO.vmec_interpolate(
        vmec.variables["bsupvmnc"][:],
        np.zeros_like(vmec.variables["bsupvmnc"][:]),
        vmec.variables["xm_nyq"][:],
        vmec.variables["xn_nyq"][:],
        theta=theta_vmec,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    bsupv_desc = VMECIO.vmec_interpolate(
        desc.variables["bsupvmnc"][:],
        np.zeros_like(desc.variables["bsupvmnc"][:]),
        desc.variables["xm_nyq"][:],
        desc.variables["xn_nyq"][:],
        theta=theta_desc,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    np.testing.assert_allclose(bsupv_vmec, bsupv_desc, rtol=1e-3)

    # B_zeta
    bsubv_vmec = VMECIO.vmec_interpolate(
        vmec.variables["bsubvmnc"][:],
        np.zeros_like(vmec.variables["bsubvmnc"][:]),
        vmec.variables["xm_nyq"][:],
        vmec.variables["xn_nyq"][:],
        theta=theta_vmec,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    bsubv_desc = VMECIO.vmec_interpolate(
        desc.variables["bsubvmnc"][:],
        np.zeros_like(desc.variables["bsubvmnc"][:]),
        desc.variables["xm_nyq"][:],
        desc.variables["xn_nyq"][:],
        theta=theta_desc,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    np.testing.assert_allclose(bsubv_vmec, bsubv_desc, rtol=1e-3)

    # straight field-line grid to compare quantities on boundary
    grid = LinearGrid(M=6, N=0, NFP=desc.variables["nfp"][:], rho=np.array([1.0]))
    theta_vmec = VMECIO.compute_theta_coords(
        vmec.variables["lmns"][:],
        vmec.variables["xm"][:],
        vmec.variables["xn"][:],
        grid.nodes[:, 0],
        grid.nodes[:, 1],
        grid.nodes[:, 2],
    )
    theta_desc = VMECIO.compute_theta_coords(
        desc.variables["lmns"][:],
        desc.variables["xm"][:],
        desc.variables["xn"][:],
        grid.nodes[:, 0],
        grid.nodes[:, 1],
        grid.nodes[:, 2],
    )

    # lambda
    L_vmec = VMECIO.vmec_interpolate(
        np.zeros_like(vmec.variables["lmns"][:]),
        vmec.variables["lmns"][:],
        vmec.variables["xm"][:],
        vmec.variables["xn"][:],
        theta=theta_vmec,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    L_desc = VMECIO.vmec_interpolate(
        np.zeros_like(desc.variables["lmns"][:]),
        desc.variables["lmns"][:],
        desc.variables["xm"][:],
        desc.variables["xn"][:],
        theta=theta_desc,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    np.testing.assert_allclose(L_vmec, L_desc, rtol=1e-2)

    # Jacobian
    g_vmec = VMECIO.vmec_interpolate(
        vmec.variables["gmnc"][:],
        np.zeros_like(vmec.variables["gmnc"][:]),
        vmec.variables["xm_nyq"][:],
        vmec.variables["xn_nyq"][:],
        theta=theta_vmec,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    g_desc = VMECIO.vmec_interpolate(
        desc.variables["gmnc"][:],
        np.zeros_like(desc.variables["gmnc"][:]),
        desc.variables["xm_nyq"][:],
        desc.variables["xn_nyq"][:],
        theta=theta_desc,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    np.testing.assert_allclose(g_vmec, g_desc, rtol=1e-2)

    # B^theta
    bsupu_vmec = VMECIO.vmec_interpolate(
        vmec.variables["bsupumnc"][:],
        np.zeros_like(vmec.variables["bsupumnc"][:]),
        vmec.variables["xm_nyq"][:],
        vmec.variables["xn_nyq"][:],
        theta=theta_vmec,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    bsupu_desc = VMECIO.vmec_interpolate(
        desc.variables["bsupumnc"][:],
        np.zeros_like(desc.variables["bsupumnc"][:]),
        desc.variables["xm_nyq"][:],
        desc.variables["xn_nyq"][:],
        theta=theta_desc,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    np.testing.assert_allclose(bsupu_vmec, bsupu_desc, rtol=1e-2)

    # B_theta
    bsubu_vmec = VMECIO.vmec_interpolate(
        vmec.variables["bsubumnc"][:],
        np.zeros_like(vmec.variables["bsubumnc"][:]),
        vmec.variables["xm_nyq"][:],
        vmec.variables["xn_nyq"][:],
        theta=theta_vmec,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    bsubu_desc = VMECIO.vmec_interpolate(
        desc.variables["bsubumnc"][:],
        np.zeros_like(desc.variables["bsubumnc"][:]),
        desc.variables["xm_nyq"][:],
        desc.variables["xn_nyq"][:],
        theta=theta_desc,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    np.testing.assert_allclose(bsubu_vmec, bsubu_desc, rtol=1e-2)

    # B_psi
    bsubs_vmec = VMECIO.vmec_interpolate(
        np.zeros_like(vmec.variables["bsubsmns"][:]),
        vmec.variables["bsubsmns"][:],
        vmec.variables["xm_nyq"][:],
        vmec.variables["xn_nyq"][:],
        theta=theta_vmec,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    bsubs_desc = VMECIO.vmec_interpolate(
        np.zeros_like(desc.variables["bsubsmns"][:]),
        desc.variables["bsubsmns"][:],
        desc.variables["xm_nyq"][:],
        desc.variables["xn_nyq"][:],
        theta=theta_desc,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    np.testing.assert_allclose(bsubs_vmec, bsubs_desc, rtol=1e-2, atol=2e-3)

    # J^theta
    curru_vmec = VMECIO.vmec_interpolate(
        vmec.variables["currumnc"][:],
        np.zeros_like(vmec.variables["currumnc"][:]),
        vmec.variables["xm_nyq"][:],
        vmec.variables["xn_nyq"][:],
        theta=theta_vmec,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    curru_desc = VMECIO.vmec_interpolate(
        desc.variables["currumnc"][:],
        np.zeros_like(desc.variables["currumnc"][:]),
        desc.variables["xm_nyq"][:],
        desc.variables["xn_nyq"][:],
        theta=theta_desc,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    np.testing.assert_allclose(curru_vmec, curru_desc, rtol=1e-2)

    # J^zeta
    currv_vmec = VMECIO.vmec_interpolate(
        vmec.variables["currvmnc"][:],
        np.zeros_like(vmec.variables["currvmnc"][:]),
        vmec.variables["xm_nyq"][:],
        vmec.variables["xn_nyq"][:],
        theta=theta_vmec,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    currv_desc = VMECIO.vmec_interpolate(
        desc.variables["currvmnc"][:],
        np.zeros_like(desc.variables["currvmnc"][:]),
        desc.variables["xm_nyq"][:],
        desc.variables["xn_nyq"][:],
        theta=theta_desc,
        phi=grid.nodes[:, 2],
        s=grid.nodes[:, 0],
        sym=False,
    )
    np.testing.assert_allclose(currv_vmec, currv_desc, rtol=1e-2)


@pytest.mark.regression
@pytest.mark.slow
def test_vmec_save_asym(VMEC_save_asym):
    """Tests that saving in NetCDF format agrees with VMEC."""
    vmec, desc, eq = VMEC_save_asym
    # first, compare some quantities which don't require calculation
    assert vmec.variables["version_"][:] == desc.variables["version_"][:]
    assert vmec.variables["mgrid_mode"][:] == desc.variables["mgrid_mode"][:]
    assert np.all(
        np.char.compare_chararrays(
            vmec.variables["mgrid_file"][:],
            desc.variables["mgrid_file"][:],
            "==",
            False,
        )
    )
    assert vmec.variables["ier_flag"][:] == desc.variables["ier_flag"][:]
    assert (
        vmec.variables["lfreeb__logical__"][:] == desc.variables["lfreeb__logical__"][:]
    )
    assert (
        vmec.variables["lrecon__logical__"][:] == desc.variables["lrecon__logical__"][:]
    )
    assert vmec.variables["lrfp__logical__"][:] == desc.variables["lrfp__logical__"][:]
    assert (
        vmec.variables["lasym__logical__"][:] == desc.variables["lasym__logical__"][:]
    )
    assert vmec.variables["nfp"][:] == desc.variables["nfp"][:]
    assert vmec.variables["ns"][:] == desc.variables["ns"][:]
    assert vmec.variables["mpol"][:] == desc.variables["mpol"][:]
    assert vmec.variables["ntor"][:] == desc.variables["ntor"][:]
    assert vmec.variables["mnmax"][:] == desc.variables["mnmax"][:]
    np.testing.assert_allclose(vmec.variables["xm"][:], desc.variables["xm"][:])
    np.testing.assert_allclose(vmec.variables["xn"][:], desc.variables["xn"][:])
    assert vmec.variables["mnmax_nyq"][:] == desc.variables["mnmax_nyq"][:]
    np.testing.assert_allclose(vmec.variables["xm_nyq"][:], desc.variables["xm_nyq"][:])
    np.testing.assert_allclose(vmec.variables["xn_nyq"][:], desc.variables["xn_nyq"][:])
    assert vmec.variables["signgs"][:] == desc.variables["signgs"][:]
    assert vmec.variables["gamma"][:] == desc.variables["gamma"][:]
    assert vmec.variables["nextcur"][:] == desc.variables["nextcur"][:]
    assert np.all(
        np.char.compare_chararrays(
            vmec.variables["pmass_type"][:],
            desc.variables["pmass_type"][:],
            "==",
            False,
        )
    )
    assert np.all(
        np.char.compare_chararrays(
            vmec.variables["piota_type"][:],
            desc.variables["piota_type"][:],
            "==",
            False,
        )
    )
    assert np.all(
        np.char.compare_chararrays(
            vmec.variables["pcurr_type"][:],
            desc.variables["pcurr_type"][:],
            "==",
            False,
        )
    )
    np.testing.assert_allclose(
        vmec.variables["am"][:], desc.variables["am"][:], atol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["ai"][:], desc.variables["ai"][:], atol=1e-8
    )
    np.testing.assert_allclose(
        vmec.variables["ac"][:], desc.variables["ac"][:], atol=3e-5
    )
    np.testing.assert_allclose(
        vmec.variables["presf"][:], desc.variables["presf"][:], atol=2e-5
    )
    np.testing.assert_allclose(vmec.variables["pres"][:], desc.variables["pres"][:])
    np.testing.assert_allclose(vmec.variables["mass"][:], desc.variables["mass"][:])
    np.testing.assert_allclose(
        vmec.variables["iotaf"][:], desc.variables["iotaf"][:], rtol=5e-4
    )
    np.testing.assert_allclose(
        vmec.variables["q_factor"][:], desc.variables["q_factor"][:], rtol=5e-4
    )
    np.testing.assert_allclose(
        vmec.variables["iotas"][:], desc.variables["iotas"][:], rtol=5e-4
    )
    np.testing.assert_allclose(vmec.variables["phi"][:], desc.variables["phi"][:])
    np.testing.assert_allclose(vmec.variables["phipf"][:], desc.variables["phipf"][:])
    np.testing.assert_allclose(vmec.variables["phips"][:], desc.variables["phips"][:])
    np.testing.assert_allclose(
        vmec.variables["chi"][:], desc.variables["chi"][:], atol=3e-5, rtol=1e-3
    )
    np.testing.assert_allclose(
        vmec.variables["chipf"][:], desc.variables["chipf"][:], atol=3e-5, rtol=1e-3
    )
    np.testing.assert_allclose(
        vmec.variables["Rmajor_p"][:], desc.variables["Rmajor_p"][:]
    )
    np.testing.assert_allclose(
        vmec.variables["Aminor_p"][:], desc.variables["Aminor_p"][:]
    )
    np.testing.assert_allclose(vmec.variables["aspect"][:], desc.variables["aspect"][:])
    np.testing.assert_allclose(
        vmec.variables["volume_p"][:], desc.variables["volume_p"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["volavgB"][:], desc.variables["volavgB"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["betatotal"][:], desc.variables["betatotal"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["betapol"][:], desc.variables["betapol"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["betator"][:], desc.variables["betator"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["ctor"][:],
        desc.variables["ctor"][:],
        atol=1e-9,  # it is a zero current solve
    )
    np.testing.assert_allclose(
        vmec.variables["rbtor"][:], desc.variables["rbtor"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["rbtor0"][:], desc.variables["rbtor0"][:], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["b0"][:], desc.variables["b0"][:], rtol=4e-3
    )
    np.testing.assert_allclose(
        vmec.variables["buco"][20:100], desc.variables["buco"][20:100], atol=1e-15
    )
    np.testing.assert_allclose(
        vmec.variables["bvco"][20:100], desc.variables["bvco"][20:100], rtol=1e-5
    )
    np.testing.assert_allclose(
        vmec.variables["vp"][20:100], desc.variables["vp"][20:100], rtol=3e-4
    )
    np.testing.assert_allclose(
        vmec.variables["bdotb"][20:100], desc.variables["bdotb"][20:100], rtol=3e-4
    )
    np.testing.assert_allclose(
        vmec.variables["jdotb"][20:100],
        desc.variables["jdotb"][20:100],
        atol=4e-3,  # nearly zero bc is vacuum
    )
    np.testing.assert_allclose(
        vmec.variables["jcuru"][20:100], desc.variables["jcuru"][20:100], atol=2
    )
    np.testing.assert_allclose(
        vmec.variables["jcurv"][20:100], desc.variables["jcurv"][20:100], rtol=2
    )
    np.testing.assert_allclose(
        vmec.variables["DShear"][20:100], desc.variables["DShear"][20:100], rtol=6e-2
    )
    np.testing.assert_allclose(
        vmec.variables["DCurr"][20:100],
        desc.variables["DCurr"][20:100],
        atol=1e-4,  # nearly zero bc vacuum
    )
    np.testing.assert_allclose(
        vmec.variables["DWell"][20:100], desc.variables["DWell"][20:100], rtol=1e-2
    )
    np.testing.assert_allclose(
        vmec.variables["DGeod"][20:100],
        desc.variables["DGeod"][20:100],
        atol=4e-3,
        rtol=1e-2,
    )

    # the Mercier stability is pretty off,
    # but these are not exactly similar solutions to eachother
    np.testing.assert_allclose(
        vmec.variables["DMerc"][20:100], desc.variables["DMerc"][20:100], atol=4e-3
    )
    np.testing.assert_allclose(
        vmec.variables["raxis_cc"][:],
        desc.variables["raxis_cc"][:],
        rtol=5e-5,
        atol=4e-3,
    )
    np.testing.assert_allclose(
        vmec.variables["zaxis_cs"][:],
        desc.variables["zaxis_cs"][:],
        rtol=5e-5,
        atol=1e-3,
    )
    np.testing.assert_allclose(
        vmec.variables["rmin_surf"][:], desc.variables["rmin_surf"][:], rtol=5e-3
    )
    np.testing.assert_allclose(
        vmec.variables["rmax_surf"][:], desc.variables["rmax_surf"][:], rtol=5e-3
    )
    np.testing.assert_allclose(
        vmec.variables["zmax_surf"][:], desc.variables["zmax_surf"][:], rtol=5e-3
    )
    np.testing.assert_allclose(
        vmec.variables["beta_vol"][:], desc.variables["beta_vol"][:], rtol=5e-5
    )
    np.testing.assert_allclose(
        vmec.variables["betaxis"][:], desc.variables["betaxis"][:], rtol=5e-5
    )
    # Next, calculate some quantities and compare
    # the DESC wout -> DESC (should be very close)
    # and the DESC wout -> VMEC wout (should be approximately close)
    surfs = desc.variables["ns"][:]
    s_full = np.linspace(0, 1, surfs)
    s_half = s_full[0:-1] + 0.5 / (surfs - 1)
    r_full = np.sqrt(s_full)
    r_half = np.sqrt(s_half)

    vol_grid = LinearGrid(
        rho=r_full[10::10],
        M=15,
        N=15,
        NFP=eq.NFP,
        axis=False,
        sym=False,
    )
    vol_half_grid = LinearGrid(
        rho=r_half[10::10],
        M=15,
        N=15,
        NFP=eq.NFP,
        axis=False,
        sym=False,
    )
    bdry_grid = LinearGrid(rho=1.0, M=15, N=15, NFP=eq.NFP, axis=False, sym=False)

    def test(
        nc_str,
        desc_str,
        negate_DESC_quant=False,
        use_nyq=True,
        convert_sqrt_g_or_B_rho=False,
        atol_desc_desc_wout=5e-5,
        rtol_desc_desc_wout=1e-5,
        atol_vmec_desc_wout=1e-5,
        rtol_vmec_desc_wout=1e-2,
        grid=vol_grid,
        is_half_grid=False,
    ):
        """Helper fxn to evaluate Fourier series from wout and compare to DESC."""
        xm = desc.variables["xm_nyq"][:] if use_nyq else desc.variables["xm"][:]
        xn = desc.variables["xn_nyq"][:] if use_nyq else desc.variables["xn"][:]

        si = np.insert(s_half, 0, 0) if is_half_grid else s_full
        rho = grid.nodes[:, 0]
        s = rho**2
        # some quantities must be negated before comparison bc
        # they are negative in the wout i.e. B^theta
        negate = -1 if negate_DESC_quant else 1

        quant_from_desc_wout = VMECIO.vmec_interpolate(
            desc.variables[nc_str + "c"][:],
            desc.variables[nc_str + "s"][:],
            xm,
            xn,
            theta=-grid.nodes[:, 1],  # -theta bc when we save wout we reverse theta
            phi=grid.nodes[:, 2],
            s=s,
            sym=False,
            si=si,
        )

        quant_from_vmec_wout = VMECIO.vmec_interpolate(
            vmec.variables[nc_str + "c"][:],
            vmec.variables[nc_str + "s"][:],
            xm,
            xn,
            # pi - theta bc VMEC, when it gets a CW angle bdry,
            # changes poloidal angle to  theta -> pi-theta
            theta=np.pi - grid.nodes[:, 1],
            phi=grid.nodes[:, 2],
            s=s,
            sym=False,
            si=si,
        )

        data = eq.compute(["rho", "sqrt(g)", desc_str], grid=grid)
        # convert sqrt(g) or B_rho->B_psi if needed
        quant_desc = (
            data[desc_str] / 2 / data["rho"]
            if convert_sqrt_g_or_B_rho
            else data[desc_str]
        )

        # add sqrt(g) factor if currents being compared
        quant_desc = (
            quant_desc * abs(data["sqrt(g)"]) / 2 / data["rho"]
            if "J" in desc_str
            else quant_desc
        )

        np.testing.assert_allclose(
            negate * quant_desc,
            quant_from_desc_wout,
            atol=atol_desc_desc_wout,
            rtol=rtol_desc_desc_wout,
        )
        np.testing.assert_allclose(
            quant_from_desc_wout,
            quant_from_vmec_wout,
            atol=atol_vmec_desc_wout,
            rtol=rtol_vmec_desc_wout,
        )

    # R & Z
    test("rmn", "R", use_nyq=False)
    test("zmn", "Z", use_nyq=False, atol_vmec_desc_wout=1e-2)
    test(
        "lmn",
        "lambda",
        use_nyq=False,
        atol_vmec_desc_wout=1e-2,
        negate_DESC_quant=True,
        grid=vol_half_grid,
        is_half_grid=True,
    )
    # |B|
    test("bmn", "|B|", rtol_desc_desc_wout=7e-4)

    # B^zeta
    test("bsupvmn", "B^zeta")

    # B_zeta
    test("bsubvmn", "B_zeta", grid=vol_half_grid, is_half_grid=True)

    # hard to compare to VMEC for the currents, since
    # VMEC F error is worse and equilibria are not exactly similar
    # just compare back to DESC
    test("currumn", "J^theta", atol_vmec_desc_wout=1e4)
    test("currvmn", "J^zeta", negate_DESC_quant=True, atol_vmec_desc_wout=1e5)

    test(
        "gmn",
        "sqrt(g)",
        convert_sqrt_g_or_B_rho=True,
        negate_DESC_quant=True,
        grid=vol_half_grid,
        is_half_grid=True,
    )

    # Compare B_psi B^theta and B_theta at bdry only
    test(
        "bsupumn",
        "B^theta",
        negate_DESC_quant=True,
        grid=bdry_grid,
        is_half_grid=True,
        atol_vmec_desc_wout=6e-4,
    )
    test(
        "bsubumn",
        "B_theta",
        negate_DESC_quant=True,
        grid=bdry_grid,
        is_half_grid=True,
    )
    test(
        "bsubsmn",
        "B_rho",
        convert_sqrt_g_or_B_rho=True,
        grid=bdry_grid,
        atol_vmec_desc_wout=2e-3,
    )


@pytest.mark.unit
@pytest.mark.mpl_image_compare(tolerance=1)
def test_plot_vmec_comparison():
    """Test that DESC and VMEC flux surface plots match."""
    eq = desc.examples.get("SOLOVEV")
    fig, ax = VMECIO.plot_vmec_comparison(eq, "tests/inputs/wout_SOLOVEV.nc")
    return fig


@pytest.mark.unit
def test_vmec_boundary_subspace(DummyStellarator):
    """Test VMEC boundary subspace is enforced properly."""
    eq = load(load_from=str(DummyStellarator["output_path"]), file_format="hdf5")

    RBC = np.array([[1, 2], [-1, 2], [1, 0], [2, 2]])
    ZBS = np.array([[2, 1], [-2, 1], [0, 2], [-1, 1]])
    opt_subspace = vmec_boundary_subspace(eq, RBC, ZBS)

    y_opt = np.arange(8)
    y = np.dot(y_opt, opt_subspace.T)

    m_desc = np.concatenate(
        (eq.surface.R_basis.modes[:, 1], eq.surface.Z_basis.modes[:, 1])
    )
    n_desc = np.concatenate(
        (eq.surface.R_basis.modes[:, 2], eq.surface.Z_basis.modes[:, 2])
    )

    m_vmec, n_vmec, zbs, rbc = ptolemy_identity_rev(m_desc, n_desc, y)

    tol = 1e-15
    rbc_ref = np.atleast_2d(np.array([0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1]))
    zbs_ref = np.atleast_2d(np.array([0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 0, 0]))
    np.testing.assert_allclose(rbc_ref, np.abs(rbc) > tol)
    np.testing.assert_allclose(zbs_ref, np.abs(zbs) > tol)


@pytest.mark.regression
@pytest.mark.solve
def test_write_vmec_input(TmpDir):
    """Test generated VMEC input file gives the original equilibrium when solved."""
    # write VMEC input file
    fname = str(TmpDir.join("input.SOLOVEV"))
    eq0 = get("SOLOVEV")
    W0 = eq0.compute("W")["W"]
    VMECIO.write_vmec_input(eq0, fname)

    # read VMEC input file and override defaults to match desc/examples/SOLOVEV
    ir = InputReader()
    inputs = ir.parse_inputs(fname)
    inputs[0]["spectral_indexing"] = "fringe"
    inputs[0]["L"] = 24

    # solve from VMEC input file
    fam = EquilibriaFamily(inputs)
    fam.solve_continuation(
        objective=inputs[0]["objective"],
        optimizer=inputs[0]["optimizer"],
        pert_order=[inp["pert_order"] for inp in inputs],
        ftol=[inp["ftol"] for inp in inputs],
        xtol=[inp["xtol"] for inp in inputs],
        gtol=[inp["gtol"] for inp in inputs],
        maxiter=[inp["maxiter"] for inp in inputs],
        verbose=2,
    )
    eq1 = fam[-1]
    W1 = eq1.compute("W")["W"]

    # check that solution matches original equilibrium
    rho_err, theta_err = area_difference_desc(eq0, eq1)
    np.testing.assert_allclose(eq0.L, eq1.L)
    np.testing.assert_allclose(eq0.M, eq1.M)
    np.testing.assert_allclose(eq0.N, eq1.N)
    np.testing.assert_allclose(eq0.NFP, eq1.NFP)
    np.testing.assert_allclose(eq0.Psi, eq1.Psi)
    np.testing.assert_allclose(eq0.p_l, eq1.p_l)
    np.testing.assert_allclose(eq0.i_l, eq1.i_l)
    np.testing.assert_allclose(eq0.Rb_lmn, eq1.Rb_lmn)
    np.testing.assert_allclose(eq0.Zb_lmn, eq1.Zb_lmn)
    np.testing.assert_allclose(W0, W1)
    np.testing.assert_allclose(rho_err, 0, atol=1e-4)
    np.testing.assert_allclose(theta_err, 0, atol=1e-4)
